# -*- coding: utf-8 -*-
from collective.easyform.api import set_actions
from collective.easyform.api import set_fields


def updateFields(obj, event):
    set_fields(obj.aq_parent, obj.schema)


def updateActions(obj, event):
    set_actions(obj.aq_parent, obj.schema)



# -*- coding: utf-8 -*-
from .validators import isTALES
from collective.easyform import config
from collective.easyform import easyformMessageFactory as _  # NOQA
from collective.easyform.config import HAS_XLSX_SUPPORT
from collective.easyform.interfaces import IAction
from plone import api
from plone.app.textfield import RichText
from plone.autoform import directives
from plone.schema import Email
from plone.supermodel.directives import fieldset
from Products.CMFPlone.utils import safe_unicode
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from z3c.form.browser.textarea import TextAreaWidget
from plone.autoform.interfaces import OMITTED_KEY
from zope.globalrequest import getRequest
from zope.i18n import translate
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory

import zope.i18nmessageid
import zope.interface
import zope.schema.interfaces


PMF = zope.i18nmessageid.MessageFactory("plone")
MODIFY_PORTAL_CONTENT = "cmf.ModifyPortalContent"


@provider(IContextAwareDefaultFactory)
def default_mail_subject(context):
    return translate(_(u"Form Submission"), context=getRequest())


def default_mail_body():
    """Default mail body for mailer action.
    Acquire 'mail_body_default.pt' or return hard coded default
    """
    try:
        portal = api.portal.get()
    except api.exc.CannotGetPortalError:
        return config.MAIL_BODY_DEFAULT

    mail_body_default = portal.restrictedTraverse(
        "easyform_mail_body_default.pt", default=None
    )
    if mail_body_default:
        return safe_unicode(mail_body_default.file.data)
    else:
        return config.MAIL_BODY_DEFAULT


class IMailer(IAction):
    """A form action adapter that will e-mail form input."""

    directives.write_permission(recipient_name=config.EDIT_ADDRESSING_PERMISSION)
    directives.read_permission(recipient_name=MODIFY_PORTAL_CONTENT)
    recipient_name = zope.schema.TextLine(
        title=_(
            u"label_formmailer_recipient_fullname", default=u"Recipient's full name"
        ),
        description=_(
            u"help_formmailer_recipient_fullname",
            default=u"The full name of the recipient of the mailed form.",
        ),
        default=u"",
        missing_value=u"",
        required=False,
    )

    directives.write_permission(recipient_email=config.EDIT_ADDRESSING_PERMISSION)
    directives.read_permission(recipient_email=MODIFY_PORTAL_CONTENT)
    recipient_email = Email(
        title=_(
            u"label_formmailer_recipient_email", default=u"Recipient's e-mail address"
        ),
        description=_(
            u"help_formmailer_recipient_email",
            default=u"The recipients e-mail address.",
        ),
        required=False,
    )
    fieldset(
        u"addressing",
        label=_("Addressing"),
        fields=["to_field", "cc_recipients", "bcc_recipients", "replyto_field"],
    )
    directives.write_permission(to_field=config.EDIT_ADDRESSING_PERMISSION)
    directives.read_permission(to_field=MODIFY_PORTAL_CONTENT)
    to_field = zope.schema.Choice(
        title=_(u"label_formmailer_to_extract", default=u"Extract Recipient From"),
        description=_(
            u"help_formmailer_to_extract",
            default=u"Choose a form field from which you wish to extract "
            u"input for the To header. If you choose anything other "
            u'than "None", this will override the "Recipient\'s " '
            u"e-mail address setting above. Be very cautious about "
            u"allowing unguarded user input for this purpose.",
        ),
        required=False,
        vocabulary="easyform.Fields",
    )

    directives.write_permission(cc_recipients=config.EDIT_ADDRESSING_PERMISSION)
    directives.read_permission(cc_recipients=MODIFY_PORTAL_CONTENT)
    cc_recipients = zope.schema.Text(
        title=_(u"label_formmailer_cc_recipients", default=u"CC Recipients"),
        description=_(
            u"help_formmailer_cc_recipients",
            default=u"E-mail addresses which receive a carbon copy.",
        ),
        default=u"",
        missing_value=u"",
        required=False,
    )

    directives.write_permission(bcc_recipients=config.EDIT_ADDRESSING_PERMISSION)
    directives.read_permission(bcc_recipients=MODIFY_PORTAL_CONTENT)
    bcc_recipients = zope.schema.Text(
        title=_(u"label_formmailer_bcc_recipients", default=u"BCC Recipients"),
        description=_(
            u"help_formmailer_bcc_recipients",
            default=u"E-mail addresses which receive a blind carbon copy.",
        ),
        default=u"",
        missing_value=u"",
        required=False,
    )

    directives.write_permission(replyto_field=config.EDIT_TECHNICAL_PERMISSION)
    directives.read_permission(replyto_field=MODIFY_PORTAL_CONTENT)
    replyto_field = zope.schema.Choice(
        title=_(u"label_formmailer_replyto_extract", default=u"Extract Reply-To From"),
        description=_(
            u"help_formmailer_replyto_extract",
            default=u"Choose a form field from which you wish to extract "
            u"input for the Reply-To header. NOTE: You should "
            u"activate e-mail address verification for the "
            u"designated field.",
        ),
        required=False,
        vocabulary="easyform.Fields",
    )
    fieldset(
        u"message",
        label=PMF("Message"),
        fields=[
            "msg_subject",
            "subject_field",
            "body_pre",
            "body_post",
            "body_footer",
            "showAll",
            "showFields",
            "includeEmpties",
            "sendCSV",
            "sendXML",
            "sendXLSX",
            "sendWithHeader",
        ],
    )
    directives.read_permission(msg_subject=MODIFY_PORTAL_CONTENT)
    msg_subject = zope.schema.TextLine(
        title=_(u"label_formmailer_subject", default=u"Subject"),
        description=_(
            u"help_formmailer_subject",
            default=u""
            u"Subject line of message. This is used if you "
            u"do not specify a subject field or if the field "
            u"is empty.",
        ),
        defaultFactory=default_mail_subject,
        missing_value=u"",
        required=False,
    )

    directives.write_permission(subject_field=config.EDIT_ADVANCED_PERMISSION)
    directives.read_permission(subject_field=MODIFY_PORTAL_CONTENT)
    subject_field = zope.schema.Choice(
        title=_(u"label_formmailer_subject_extract", default=u"Extract Subject From"),
        description=_(
            u"help_formmailer_subject_extract",
            default=u""
            u"Choose a form field from which you wish to extract "
            u"input for the mail subject line.",
        ),
        required=False,
        vocabulary="easyform.Fields",
    )
    directives.read_permission(body_pre=MODIFY_PORTAL_CONTENT)
    directives.widget("body_pre", TextAreaWidget)

    body_pre = RichText(
        title=_(u"label_formmailer_body_pre", default=u"Body (prepended)"),
        description=_(
            u"help_formmailer_body_pre",
            default=u"Text prepended to fields listed in mail-body",
        ),
        default=u"",
        missing_value=u"",
        default_mime_type="text/x-web-intelligent",
        allowed_mime_types=("text/x-web-intelligent",),
        output_mime_type="text/x-html-safe",
        required=False,
    )
    directives.read_permission(body_post=MODIFY_PORTAL_CONTENT)
    directives.widget("body_post", TextAreaWidget)
    body_post = RichText(
        title=_(u"label_formmailer_body_post", default=u"Body (appended)"),
        description=_(
            u"help_formmailer_body_post",
            default=u"Text appended to fields listed in mail-body",
        ),
        default=u"",
        missing_value=u"",
        default_mime_type="text/x-web-intelligent",
        allowed_mime_types=("text/x-web-intelligent",),
        output_mime_type="text/x-html-safe",
        required=False,
    )
    directives.read_permission(body_footer=MODIFY_PORTAL_CONTENT)
    directives.widget("body_footer", TextAreaWidget)
    body_footer = RichText(
        title=_(u"label_formmailer_body_footer", default=u"Body (signature)"),
        description=_(
            u"help_formmailer_body_footer",
            default=u"Text used as the footer at "
            u"bottom, delimited from the body by a dashed line.",
        ),
        default=u"",
        missing_value=u"",
        default_mime_type="text/x-web-intelligent",
        allowed_mime_types=("text/x-web-intelligent",),
        output_mime_type="text/x-html-safe",
        required=False,
    )

    directives.read_permission(showAll=MODIFY_PORTAL_CONTENT)
    showAll = zope.schema.Bool(
        title=_(u"label_mailallfields_text", default=u"Include All Fields"),
        description=_(
            u"help_mailallfields_text",
            default=u""
            u"Check this to include input for all fields "
            u"(except label and file fields). If you check "
            u"this, the choices in the pick box below "
            u"will be ignored.",
        ),
        default=True,
        required=False,
    )

    directives.read_permission(showFields=MODIFY_PORTAL_CONTENT)
    showFields = zope.schema.List(
        title=_(u"label_mailfields_text", default=u"Show Responses"),
        description=_(
            u"help_mailfields_text",
            default=u"Pick the fields whose inputs you'd like to include in "
            u"the e-mail.",
        ),
        unique=True,
        required=False,
        value_type=zope.schema.Choice(vocabulary="easyform.Fields"),
    )

    directives.read_permission(includeEmpties=MODIFY_PORTAL_CONTENT)
    includeEmpties = zope.schema.Bool(
        title=_(u"label_mailEmpties_text", default=u"Include Empties"),
        description=_(
            u"help_mailEmpties_text",
            default=u""
            u"Check this to include titles "
            u"for fields that received no input. Uncheck "
            u"to leave fields with no input out of the e-mail.",
        ),
        default=True,
        required=False,
    )

    directives.read_permission(sendCSV=MODIFY_PORTAL_CONTENT)
    sendCSV = zope.schema.Bool(
        title=_(u"label_sendCSV_text", default=u"Send CSV data attachment"),
        description=_(
            u"help_sendCSV_text",
            default=u""
            u"Check this to send a CSV file "
            u"attachment containing the values "
            u"filled out in the form.",
        ),
        default=False,
        required=False,
    )

    directives.read_permission(sendWithHeader=MODIFY_PORTAL_CONTENT)
    sendWithHeader = zope.schema.Bool(
        title=_(u"label_sendWithHeader_text", default=u"Include header in attached CSV/XLSX data"),
        description=_(
            u"help_sendWithHeader_text",
            default=u""
            u"Check this to include the CSV/XLSX header "
            u"in file attachments.",
        ),
        default=False,
        required=False,
    )

    directives.read_permission(sendXLSX=MODIFY_PORTAL_CONTENT)
    sendXLSX = zope.schema.Bool(
        title=_(u"label_sendXLSX_text", default=u"Send XLSX data attachment"),
        description=_(
            u"help_sendXLSX_text",
            default=u""
            u"Check this to send a XLSX file "
            u"attachment containing the values "
            u"filled out in the form.",
        ),
        default=False,
        required=False,
    )


    directives.read_permission(sendXML=MODIFY_PORTAL_CONTENT)
    sendXML = zope.schema.Bool(
        title=_(u"label_sendXML_text", default=u"Send XML data attachment"),
        description=_(
            u"help_sendXML_text",
            default=u""
            u"Check this to send an XML file "
            u"attachment containing the values "
            u"filled out in the form.",
        ),
        default=False,
        required=False,
    )

    fieldset(u"template", label=PMF("Template"), fields=["body_pt", "body_type"])
    directives.write_permission(body_pt=config.EDIT_TALES_PERMISSION)
    directives.read_permission(body_pt=MODIFY_PORTAL_CONTENT)
    body_pt = zope.schema.Text(
        title=_(u"label_formmailer_body_pt", default=u"Mail-Body Template"),
        description=_(
            u"help_formmailer_body_pt",
            default=u"This is a Zope Page Template used for rendering of "
            u"the mail-body. You don't need to modify it, but if you "
            u"know TAL (Zope's Template Attribute Language) have "
            u"the full power to customize your outgoing mails.",
        ),
        defaultFactory=default_mail_body,
        missing_value=u"",
    )

    directives.write_permission(body_type=config.EDIT_ADVANCED_PERMISSION)
    directives.read_permission(body_type=MODIFY_PORTAL_CONTENT)
    body_type = zope.schema.Choice(
        title=_(u"label_formmailer_body_type", default=u"Mail Format"),
        description=_(
            u"help_formmailer_body_type",
            default=u"Set the mime-type of the mail-body. Change this "
            u"setting only if you know exactly what you are doing. "
            u"Leave it blank for default behaviour.",
        ),
        default=u"html",
        vocabulary="easyform.MimeList",
    )
    fieldset(
        u"headers", label=_("Headers"), fields=["xinfo_headers", "additional_headers"]
    )
    directives.widget(xinfo_headers=CheckBoxFieldWidget)
    directives.write_permission(xinfo_headers=config.EDIT_TECHNICAL_PERMISSION)
    directives.read_permission(xinfo_headers=MODIFY_PORTAL_CONTENT)
    xinfo_headers = zope.schema.List(
        title=_(u"label_xinfo_headers_text", default=u"HTTP Headers"),
        description=_(
            u"help_xinfo_headers_text",
            default=u""
            u"Pick any items from the HTTP headers that "
            u"you'd like to insert as X- headers in the message.",
        ),
        unique=True,
        required=False,
        default=[u"HTTP_X_FORWARDED_FOR", u"REMOTE_ADDR", u"PATH_INFO"],
        missing_value=[u"HTTP_X_FORWARDED_FOR", u"REMOTE_ADDR", u"PATH_INFO"],
        value_type=zope.schema.Choice(vocabulary="easyform.XinfoHeaders"),
    )
    directives.write_permission(additional_headers=config.EDIT_TECHNICAL_PERMISSION)
    directives.read_permission(additional_headers=MODIFY_PORTAL_CONTENT)
    additional_headers = zope.schema.List(
        title=_(u"label_formmailer_additional_headers", default=u"Additional Headers"),
        description=_(
            u"help_formmailer_additional_headers",
            default=u"Additional e-mail-header lines. Only use "
            u"RFC822-compliant headers.",
        ),
        unique=True,
        required=False,
        value_type=zope.schema.TextLine(
            title=_(
                u"extra_header", default=u"${name} Header", mapping={u"name": u"HTTP"}
            )
        ),
    )

    fieldset(
        u"overrides",
        label=_("Overrides"),
        fields=[
            "subjectOverride",
            "senderOverride",
            "recipientOverride",
            "ccOverride",
            "bccOverride",
        ],
    )

    directives.write_permission(subjectOverride=config.EDIT_TALES_PERMISSION)
    directives.read_permission(subjectOverride=MODIFY_PORTAL_CONTENT)
    subjectOverride = zope.schema.TextLine(
        title=_(u"label_subject_override_text", default=u"Subject Expression"),
        description=_(
            u"help_subject_override_text",
            default=u"A TALES expression that will be evaluated to override "
            u"any value otherwise entered for the e-mail subject "
            u"header. Leave empty if unneeded. Your expression "
            u"should evaluate as a string. PLEASE NOTE: errors in "
            u"the evaluation of this expression will cause an error "
            u"on form display.",
        ),
        required=False,
        default=u"",
        missing_value=u"",
        constraint=isTALES,
    )

    directives.write_permission(senderOverride=config.EDIT_TALES_PERMISSION)
    directives.read_permission(senderOverride=MODIFY_PORTAL_CONTENT)
    senderOverride = zope.schema.TextLine(
        title=_(u"label_sender_override_text", default=u"Sender Expression"),
        description=_(
            u"help_sender_override_text",
            default=u"A TALES expression that will be evaluated to override "
            u'the "From" header. Leave empty if unneeded. '
            u"Your expression should evaluate as a string. "
            u"Example: python:fields['replyto'] "
            u"PLEASE NOTE: errors in the evaluation of this "
            u"expression will cause an error on form display.",
        ),
        required=False,
        default=u"",
        missing_value=u"",
        constraint=isTALES,
    )

    directives.write_permission(recipientOverride=config.EDIT_TALES_PERMISSION)
    directives.read_permission(recipientOverride=MODIFY_PORTAL_CONTENT)
    recipientOverride = zope.schema.TextLine(
        title=_(u"label_recipient_override_text", default=u"Recipient Expression"),
        description=_(
            u"help_recipient_override_text",
            default=u"A TALES expression that will be evaluated to override "
            u"any value otherwise entered for the recipient "
            u"e-mail address. You are strongly cautioned against using"
            u"unvalidated data from the request for this purpose. "
            u"Leave empty if unneeded. Your expression should "
            u"evaluate as a string. PLEASE NOTE: errors in the "
            u"evaluation of this expression will cause "
            u"an error on form display.",
        ),
        required=False,
        default=u"",
        missing_value=u"",
        constraint=isTALES,
    )

    directives.write_permission(ccOverride=config.EDIT_TALES_PERMISSION)
    directives.read_permission(ccOverride=MODIFY_PORTAL_CONTENT)
    ccOverride = zope.schema.TextLine(
        title=_(u"label_cc_override_text", default=u"CC Expression"),
        description=_(
            u"help_cc_override_text",
            default=u"A TALES expression that will be evaluated to override "
            u"any value otherwise entered for the CC list. You are "
            u"strongly cautioned against using unvalidated data from "
            u"the request for this purpose. Leave empty if unneeded. "
            u"Your expression should evaluate as a sequence of "
            u"strings. PLEASE NOTE: errors in the evaluation of this "
            u"expression will cause an error on form display.",
        ),
        required=False,
        default=u"",
        missing_value=u"",
        constraint=isTALES,
    )

    directives.write_permission(bccOverride=config.EDIT_TALES_PERMISSION)
    directives.read_permission(bccOverride=MODIFY_PORTAL_CONTENT)
    bccOverride = zope.schema.TextLine(
        title=_(u"label_bcc_override_text", default=u"BCC Expression"),
        description=_(
            u"help_bcc_override_text",
            default=u"A TALES expression that will be evaluated to override "
            u"any value otherwise entered for the BCC list. "
            u"You are strongly cautioned against using "
            u"unvalidated data from the request for this purpose. "
            u"Leave empty if unneeded. Your expression should "
            u"evaluate as a sequence of strings. PLEASE NOTE: errors "
            u"in the evaluation of this expression will cause "
            u"an error on form display.",
        ),
        required=False,
        default=u"",
        missing_value=u"",
        constraint=isTALES,
    )


# extend list of omitted fields if XLSX extra is not available
if not HAS_XLSX_SUPPORT:
    omitted_fields = IMailer.queryTaggedValue(OMITTED_KEY, [])[:]
    omitted_fields.append((zope.interface.Interface, 'sendXLSX', 'true'))
    IMailer.setTaggedValue(OMITTED_KEY, omitted_fields)




# -*- coding: utf-8 -*-
from collective.easyform.interfaces import IEasyForm
from plone.dexterity.content import Item
from zope.interface import implementer


@implementer(IEasyForm)
class EasyForm(Item):
    """An easy form content base"""



import unittest
import six

from collective.easyform.tests.base import EasyFormFunctionalTestCase


class LikertFieldTests(unittest.TestCase):

    def _getTargetClass(self):
        from collective.easyform.fields import Likert
        return Likert

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_ctor_defaults(self):
        txt = self._makeOne()
        self.assertEqual(txt._type, six.text_type)

    def test_validate_not_required(self):
        field = self._makeOne(required=False)
        field.validate(None)
        field.validate(u'')

    def test_validate_with_answers(self):
        field = self._makeOne(required=False, questions=[u'Question 1'], answers=[u'Agree', u'Disagree'])
        field.validate(None)
        field.validate(u'')
        field.validate(u'1: Agree')
        self.assertRaises(ValueError, field.validate, u'1:agree')
        self.assertRaises(ValueError, field.validate, u'-1:agree')
        self.assertRaises(ValueError, field.validate, u'Agree')

    def test_validate_with_more_answers(self):
        field = self._makeOne(required=False, questions=[u'Question 1', u'Question 2'], answers=[u'Agree', u'Disagree'])
        field.validate(None)
        field.validate(u'')
        field.validate(u'1: Agree')
        field.validate(u'2: Agree')
        field.validate(u'1: Disagree, 2: Agree')
        self.assertRaises(ValueError, field.validate, u'1:agree')
        self.assertRaises(ValueError, field.validate, u'-1:agree')
        self.assertRaises(ValueError, field.validate, u'Agree')

    def test_parse(self):
        field = self._makeOne(required=False, questions=[u'Question 1', u'Question 2'], answers=[u'Agree', u'Disagree'])
        field.validate(None)
        self.assertEquals(dict(), field.parse(u''))
        self.assertEquals({1: u'Agree'}, field.parse(u'1: Agree'))
        self.assertEquals({2: u'Agree'}, field.parse(u'2: Agree'))
        self.assertEquals(
            {1: u'Disagree', 2: u'Agree'},
            field.parse(u'1: Disagree, 2: Agree')
        )


class LikerWidgetTests(EasyFormFunctionalTestCase):

    def setUp(self):
        super(LikerWidgetTests, self).setUp()
        ff1 = getattr(self.folder, "ff1")
        self.assertEqual(ff1.portal_type, u'EasyForm')

        from zope.interface import Interface
        from collective.easyform.api import set_fields
        from collective.easyform.api import set_actions
        from collective.easyform.actions import SaveData
        from collective.easyform.fields import Likert

        class IWithLikert(Interface):
            likert = Likert(questions=[u"Q1", u"Q2"], answers=[u"Agree", u"Disagree"])

        set_fields(ff1, IWithLikert)

        class IWithSaver(Interface):
            saver = SaveData(showFields=[])

        IWithSaver.setTaggedValue('context', ff1)
        set_actions(ff1, IWithSaver)
        import transaction
        transaction.commit()

    def test_widget_input(self):
        from zope.schema import getFieldsInOrder
        from collective.easyform.api import get_schema
        from collective.easyform.fields import Likert

        ff1 = getattr(self.folder, "ff1")
        schema = get_schema(ff1)
        fields = getFieldsInOrder(schema)
        self.assertEqual(len(fields), 1)
        self.assertEqual(fields[0][0], 'likert')
        self.assertTrue(isinstance(fields[0][1], Likert))

        # check that LikertWidget is used
        # and that questions and answers are rendered in a table
        view = ff1.restrictedTraverse('view')
        rendered = view()
        self.assertTrue(u"likert-widget" in rendered)
        self.assertTrue(u"<span>Q1</span>" in rendered)
        self.assertTrue(u"<span>Q2</span>" in rendered)
        self.assertTrue(u"<th>Agree</th>" in rendered)
        self.assertTrue(u"<th>Disagree</th>" in rendered)

    def test_likert_saved(self):
        import transaction
        ff1 = getattr(self.folder, "ff1")
        view = ff1.restrictedTraverse('view')
        # submit data to test value extraction from widget
        ff1.CSRFProtection = False  # no csrf protection
        view.request.form = {
            "form.buttons.submit": "Send",
            "form.widgets.likert.0": "Agree",
            "form.widgets.likert.1": "Disagree",
        }
        view.request.method = 'POST'
        rendered = view()
        self.assertTrue(u"Thank You" in rendered)
        self.assertTrue(u"1: Agree, 2: Disagree" in rendered)
        transaction.commit()
        actions_view = ff1.restrictedTraverse('actions')
        saver_view = actions_view.publishTraverse(actions_view.request, 'saver')
        data_view = saver_view.restrictedTraverse('@@data')
        rendered = data_view()
        from bs4 import BeautifulSoup
        radio_buttons = BeautifulSoup(rendered, 'html.parser').find_all(type="radio")
        self.assertEquals(len(radio_buttons), 4) # 2 rows of 2 answers

        # First question: answered Agree -> Agree input checked
        self.assertTrue("0_0" in radio_buttons[0]['id'])
        self.assertEquals(radio_buttons[0]['value'], "Agree")
        self.assertTrue(radio_buttons[0].has_attr('checked'))

        # First question: answered Agree -> Disagree input not checked
        self.assertTrue("0_1" in radio_buttons[1]['id'])
        self.assertEquals(radio_buttons[1]['value'], "Disagree")
        self.assertFalse(radio_buttons[1].has_attr('checked'))

        # Second question: answered Disagree -> Agree input not checked
        self.assertTrue("1_0" in radio_buttons[2]['id'])
        self.assertEquals(radio_buttons[2]['value'], "Agree")
        self.assertFalse(radio_buttons[2].has_attr('checked'))

        # Second question: answered Disagree -> Disagree input checked
        self.assertTrue("1_1" in radio_buttons[3]['id'])
        self.assertEquals(radio_buttons[3]['value'], "Disagree")
        self.assertTrue(radio_buttons[3].has_attr('checked'))



# -*- coding: utf-8 -*-
from .actions import IAction
from collective.easyform import config
from collective.easyform import easyformMessageFactory as _
from plone.autoform import directives

import zope.interface
import zope.schema.interfaces


MODIFY_PORTAL_CONTENT = "cmf.ModifyPortalContent"


class ICustomScript(IAction):
    """Executes a Python script for form data."""

    directives.read_permission(ProxyRole=MODIFY_PORTAL_CONTENT)
    directives.write_permission(ProxyRole=config.EDIT_PYTHON_PERMISSION)
    ProxyRole = zope.schema.Choice(
        title=_(u"label_script_proxy", default=u"Proxy role"),
        description=_(
            u"help_script_proxy", default=u"Role under which to run the script."
        ),
        default=u"none",
        required=True,
        vocabulary="easyform.ProxyRoleChoices",
    )
    directives.read_permission(ScriptBody=MODIFY_PORTAL_CONTENT)
    directives.write_permission(ScriptBody=config.EDIT_PYTHON_PERMISSION)
    ScriptBody = zope.schema.Text(
        title=_(u"label_script_body", default=u"Script body"),
        description=_(u"help_script_body", default=u"Write your script here."),
        default=config.DEFAULT_SCRIPT,
        required=False,
        missing_value=u"",
    )



# -*- coding: utf-8 -*-
from collections import namedtuple
from collective.easyform.migration.actions import actions_model
from collective.easyform.migration.data import migrate_saved_data
from collective.easyform.migration.fields import fields_model
from plone import api
from plone.app.contenttypes.migration.field_migrators import (  # noqa
    migrate_richtextfield,
)
from plone.app.contenttypes.migration.field_migrators import migrate_simplefield  # noqa
from plone.app.contenttypes.migration.migration import ATCTContentMigrator
from plone.app.contenttypes.migration.migration import migrate
from plone.autoform.form import AutoExtensibleForm
from plone.protect.interfaces import IDisableCSRFProtection
from plone.supermodel import model
from six import StringIO
from z3c.form.button import buttonAndHandler
from z3c.form.form import Form
from zope import schema
from zope.interface import alsoProvides

import logging
import transaction


logger = logging.getLogger("collective.easyform.migration")

Field = namedtuple("Type", ["name", "handler"])

FIELD_MAPPING = {
    "submitLabel": Field("submitLabel", migrate_simplefield),
    "resetLabel": Field("resetLabel", migrate_simplefield),
    "useCancelButton": Field("useCancelButton", migrate_simplefield),
    "forceSSL": Field("forceSSL", migrate_simplefield),
    "formPrologue": Field("formPrologue", migrate_richtextfield),
    "formEpilogue": Field("formEpilogue", migrate_richtextfield),
    "thanksPageOverride": Field("thanksPageOverride", migrate_simplefield),
    "formActionOverride": Field("formActionOverride", migrate_simplefield),
    "onDisplayOverride": Field("onDisplayOverride", migrate_simplefield),
    "afterValidationOverride": Field(
        "afterValidationOverride", migrate_simplefield
    ),  # noqa
    "headerInjection": Field("headerInjection", migrate_simplefield),
    "checkAuthenticator": Field("CSRFProtection", migrate_simplefield),
}


class PloneFormGenMigrator(ATCTContentMigrator):
    """Migrator for PFG to easyform"""

    src_portal_type = "FormFolder"
    src_meta_type = "FormFolder"
    dst_portal_type = "EasyForm"
    dst_meta_type = None  # not used

    def migrate_ploneformgen(self):
        for pfg_field, ef_field in FIELD_MAPPING.items():
            ef_field.handler(self.old, self.new, pfg_field, ef_field.name)
        self.new.fields_model = fields_model(self.old)
        self.new.actions_model = actions_model(self.old)

        migrate_saved_data(self.old, self.new)

    def migrate(self, unittest=0):
        super(PloneFormGenMigrator, self).migrate()
        logger.info("Migrated FormFolder %s", "/".join(self.new.getPhysicalPath()))


class IMigratePloneFormGenFormSchema(model.Schema):
    dry_run = schema.Bool(
        title=u"Dry run",
        required=True,
        default=False,
    )


class MigratePloneFormGenForm(AutoExtensibleForm, Form):
    label = u"Migrate PloneFormGen Forms"
    ignoreContext = True
    schema = IMigratePloneFormGenFormSchema

    @buttonAndHandler(u"Migrate")
    def handle_migrate(self, action):
        data, errors = self.extractData()
        if len(errors) > 0:
            return

        self.log = StringIO()
        handler = logging.StreamHandler(self.log)
        logger.addHandler(handler)
        formatter = logging.Formatter("%(asctime)s %(levelname)s %(name)s %(message)s")
        handler.setFormatter(formatter)

        self.migrate()

        self.migration_done = True
        if data.get("dry_run", False):
            transaction.abort()
            logger.info(u"PloneFormGen migration finished (dry run)")
        else:
            logger.info(u"PloneFormGen migration finished")

    def migrate(self):
        alsoProvides(self.request, IDisableCSRFProtection)
        portal = api.portal.get()
        migrate(portal, PloneFormGenMigrator)

    def render(self):
        if getattr(self, "migration_done", False):
            return self.log.getvalue()
        return super(MigratePloneFormGenForm, self).render()



__docformat__ = "reStructuredText"
import zope.component
import zope.interface
import zope.schema.interfaces
from zope.pagetemplate.interfaces import IPageTemplate
from plone.memoize.view import memoize

from z3c.form import interfaces
from z3c.form.widget import Widget, FieldWidget
from z3c.form.browser import widget

from collective.easyform.interfaces import ILikertWidget

@zope.interface.implementer_only(ILikertWidget)
class LikertWidget(widget.HTMLTextInputWidget, Widget):
    """Input type text widget implementation."""

    klass = u'likert-widget'
    css = u'text'
    value = u''

    def update(self):
        super(LikertWidget, self).update()
        widget.addFieldClass(self)

    def extract(self, default=interfaces.NO_VALUE):
        """See z3c.form.interfaces.IWidget."""
        answers = []
        if self.field.questions is None:
            return ''
        for index, question in enumerate(self.field.questions):
            question_answer = self.extract_question_answer(index, default)
            if question_answer is not None:
                answers.append(question_answer)
        return u', '.join(answers)

    def extract_question_answer(self, index, default):
        """See z3c.form.interfaces.IWidget."""
        name = '%s.%i' % (self.name, index)
        if (name not in self.request and
            '%s-empty-marker' % name in self.request):
            return None
        value = self.request.get(name, default)
        if value != default:
            return '%i: %s' % (index + 1, value)
        else:
            return None

    @memoize
    def parsed_values(self):
        return self.field.parse(self.value)

    def checked(self, question_number, answer_number):
        values = self.parsed_values()
        if not values or (question_number) not in values:
            return False
        else:
            return values[question_number] == self.field.answers[answer_number - 1]


@zope.component.adapter(zope.schema.interfaces.IField, interfaces.IFormLayer)
@zope.interface.implementer(interfaces.IFieldWidget)
def LikertFieldWidget(field, request):
    """IFieldWidget factory for TextWidget."""
    return FieldWidget(field, LikertWidget(request))



# -*- coding: utf-8 -*-

from .actions import IAction  # noqa
from .actions import IActionEditForm  # noqa
from .actions import IActionExtender  # noqa
from .actions import IActionFactory  # noqa
from .actions import IEasyFormActionContext  # noqa
from .actions import IEasyFormActionsContext  # noqa
from .actions import IEasyFormActionsEditorExtender  # noqa
from .actions import INewAction  # noqa
from .customscript import ICustomScript  # noqa
from .easyform import IEasyForm  # noqa
from .easyform import IEasyFormImportFormSchema  # noqa
from .easyform import IEasyFormThanksPage  # noqa
from .easyform import IRenderWidget  # noqa
from .fields import IEasyFormFieldContext  # noqa
from .fields import IEasyFormFieldsContext  # noqa
from .fields import IEasyFormFieldsEditorExtender  # noqa
from .fields import IEasyFormWidget  # noqa
from .fields import IFieldExtender  # noqa
from .fields import IFieldValidator  # noqa
from .fields import ILabel  # noqa
from .fields import ILabelWidget  # noqa
from .fields import INorobotCaptcha  # noqa
from .fields import IReCaptcha  # noqa
from .fields import ILikert  # noqa
from .fields import ILikertWidget  # noqa
from .fields import IHCaptcha  # noqa
from .fields import IRichLabel  # noqa
from .fields import IRichLabelWidget  # noqa
from .layer import IEasyFormLayer  # noqa
from .mailer import IMailer  # noqa
from .savedata import IExtraData  # noqa
from .savedata import ISaveData  # noqa
from .savedata import ISavedDataFormWrapper  # noqa
from zope.interface import Interface


class IEasyFormView(Interface):

    """
    EasyForm view interface
    """


class IEasyFormForm(Interface):

    """
    EasyForm view interface
    """



# -*- coding: utf-8 -*-
try:
    from StringIO import StringIO  # for Python 2
except ImportError:
    from io import StringIO  # for Python 3
from collective.easyform import validators
from collective.easyform.api import get_schema
from collective.easyform.api import set_fields
from collective.easyform.browser.view import EasyFormForm
from collective.easyform.browser.view import ValidateFile
from collective.easyform.interfaces import IFieldExtender
from collective.easyform.tests import base
from os.path import dirname
from os.path import join
from plone import api

from plone.namedfile.file import NamedFile
from plone.namedfile.interfaces import INamed
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.RegistrationTool import EmailAddressInvalid
from Products.validation import validation
from z3c.form.interfaces import IFormLayer
from zope.component import getUtility
from zope.interface.interfaces import ComponentLookupError
from zope.i18n import translate
from zope.interface import classImplements
from ZPublisher.BaseRequest import BaseRequest
from ZPublisher.HTTPRequest import FileUpload

import unittest

try:
    from plone.formwidget.hcaptcha.interfaces import IHCaptchaSettings
    HAS_HCAPTCHA = True
except ImportError:
    HAS_HCAPTCHA = False

try:
    from plone.formwidget.recaptcha.interfaces import IReCaptchaSettings
    HAS_RECAPTCHA = True
except ImportError:
    HAS_RECAPTCHA = False

IFieldValidator = validators.IFieldValidator

FORM_DATA = {
    "topic": u"test subject",
    "replyto": u"test@test.org",
    "comments": u"test comments",
}


class TestBaseValidators(base.EasyFormTestCase):

    """test base validators"""

    def afterSetUp(self):
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")
        self.ff1.CSRFProtection = False  # no csrf protection
        classImplements(BaseRequest, IFormLayer)
        validators.update_validators()

        request = self.layer["request"]
        for i in FORM_DATA:
            request.form["form.widgets.{0}".format(i)] = FORM_DATA[i]

    def test_defaultvalidator(self):
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(errors, ())
        self.assertEqual(data, FORM_DATA)

    def test_basevalidator(self):
        fields = get_schema(self.ff1)
        IFieldExtender(fields["replyto"]).validators = ["isEmail"]
        set_fields(self.ff1, fields)
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(errors, ())
        self.assertEqual(data, FORM_DATA)

    def test_basevalidator2(self):
        fields = get_schema(self.ff1)
        IFieldExtender(fields["comments"]).validators = ["isInt", "isURL"]
        set_fields(self.ff1, fields)
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(len(errors), 1)

    # Test value None and not required
    def test_basevalidator3(self):

        request = self.layer["request"]
        request.form["form.widgets.replyto"] = None

        fields = get_schema(self.ff1)

        fields["replyto"].required = False
        IFieldExtender(fields["replyto"]).validators = ["isEmail"]
        set_fields(self.ff1, fields)
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(errors, ())

    # Test value None and required
    def test_basevalidator4(self):

        request = self.layer["request"]
        request.form["form.widgets.replyto"] = None

        fields = get_schema(self.ff1)

        IFieldExtender(fields["replyto"]).validators = ["isEmail"]
        set_fields(self.ff1, fields)
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(len(errors), 1)

    def test_talvalidator(self):
        fields = get_schema(self.ff1)
        IFieldExtender(fields["comments"]).TValidator = "python: value == 'comments'"
        set_fields(self.ff1, fields)
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(errors, ())
        self.assertEqual(data, FORM_DATA)

    def test_talvalidator2(self):
        fields = get_schema(self.ff1)
        IFieldExtender(fields["comments"]).TValidator = "python: !value"
        set_fields(self.ff1, fields)
        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(len(errors), 1)

    def test_validator_translation(self):
        request = self.layer["request"]
        request["LANGUAGE"] = "de"
        request.LANGUAGE_TOOL.LANGUAGE = "de"

        fields = get_schema(self.ff1)
        IFieldExtender(fields["comments"]).validators = ["isInternationalPhoneNumber"]
        set_fields(self.ff1, fields)

        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance
        form.update()

        data, errors = form.extractData()
        self.assertEqual(len(errors), 1)

        expected_error_message = (
            u"Validierung fehlgeschlagen (isInternationalPhoneNumber): "
            u"'testcomments' Ist keine g\xfcltige internationale Telefonnummer"
        )
        self.assertEqual(errors[0].createMessage(), expected_error_message)


class LoadFixtureBase(base.EasyFormTestCase):
    """test validator in form outside of fieldset

    The test methods are reused in TestFieldsetValidator.
    They use the same field, except that one has it in a fieldset.
    """

    schema_fixture = "single_field.xml"

    def afterSetUp(self):
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")
        self.ff1.CSRFProtection = False  # no csrf protection
        self.ff1.showAll = True
        field_template = api.content.create(
            self.layer["portal"], "File", id="easyform_default_fields.xml"
        )
        with open(join(dirname(__file__), "fixtures", self.schema_fixture)) as f:
            filecontent = NamedFile(f.read(), contentType="application/xml")
        field_template.file = filecontent
        classImplements(BaseRequest, IFormLayer)
        validators.update_validators()

    def LoadRequestForm(self, **kwargs):
        request = self.layer["request"]
        request.form.clear()
        prefix = "form.widgets."
        for key in kwargs.keys():
            request.form[prefix + key] = kwargs[key]
        return request


class TestSingleFieldValidator(LoadFixtureBase):

    """test validator in form outside of fieldset

    The test methods are reused in TestFieldsetValidator.
    They use the same field, except that one has it in a fieldset.
    """

    schema_fixture = "single_field.xml"

    def test_get_default(self):
        # With a GET, we should see the default value in the form.
        request = self.LoadRequestForm()
        request.method = "GET"
        form = EasyFormForm(self.ff1, request)()
        self.assertNotIn("Required input is missing.", form)
        self.assertIn('value="foo@example.org"', form)

    def test_required(self):
        data = {"replyto": ""}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertIn("Required input is missing.", form)
        self.assertNotIn("Invalid email address.", form)

    def test_validator_in_fieldset(self):
        data = {
            "replyto": "bad email address",
        }
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertNotIn("Required input is missing.", form)
        self.assertIn("Invalid email address.", form)


class TestFieldsetValidator(TestSingleFieldValidator):

    """test validator in fieldset

    This reuses the test methods from TestSingleFieldValidator.
    """

    schema_fixture = "fieldset_with_single_field.xml"


class TestCustomValidators(base.EasyFormTestCase):
    """test our validators"""

    def ttest_inExNumericRange(self):
        v = validation.validatorFor("inExNumericRange")
        self.assertEqual(v(10, minval=1, maxval=20), 1)
        self.assertEqual(v("10", minval=1, maxval=20), 1)
        self.assertEqual(v("1", minval=1, maxval=20), 1)
        self.assertEqual(v("20", minval=1, maxval=20), 1)
        self.assertNotEqual(v(0, minval=1, maxval=5), 1)
        self.assertNotEqual(v(6, minval=1, maxval=5), 1)
        self.assertNotEqual(v(4, minval=5, maxval=3), 1)

    def ttest_isNotTooLong(self):
        v = validation.validatorFor("isNotTooLong")
        self.assertEqual(v("", maxlength=20), 1)
        self.assertEqual(v("1234567890", maxlength=20), 1)
        self.assertEqual(v("1234567890", maxlength=10), 1)
        self.assertEqual(v("1234567890", maxlength=0), 1)
        self.assertNotEqual(v("1234567890", maxlength=9), 1)
        self.assertNotEqual(v("1234567890", maxlength=1), 1)

    def test_isChecked(self):
        v = getUtility(IFieldValidator, name="isChecked")
        self.assertEqual(v("1"), None)
        self.assertNotEqual(v("0"), None)

    def test_isUnchecked(self):
        v = getUtility(IFieldValidator, name="isUnchecked")
        self.assertEqual(v("0"), None)
        self.assertNotEqual(v("1"), None)

    def test_isNotLinkSpam(self):
        v = getUtility(IFieldValidator, name="isNotLinkSpam")
        good = """I am link free and proud of it"""
        bad1 = """<a href="mylink">Bad.</a>"""
        bad2 = """http://bad.com"""
        bad3 = """www.Bad.com"""
        bad = (bad1, bad2, bad3)

        self.assertEqual(v(good), None)
        for b in bad:
            self.assertNotEqual(
                v(b), None, '"{0}" should be considered a link.'.format(b)
            )

    def ttest_isNotTooLong2(self):
        v = validation.validatorFor("isNotTooLong")
        v.maxlength = 10
        self.assertEqual(v("abc"), 1)
        self.assertNotEqual(v("abcdefghijklmnopqrstuvwxyz"), 1)

        # there was a bug where widget.maxlength could possibly be defined as
        # '' which means calling int(widget.maxlength) would fail

        class Mock(object):
            pass

        field = Mock()
        field.widget = Mock()
        field.widget.maxlength = ""

        self.assertEqual(v("abc", field=field), 1)

    def test_isEmail(self):
        v = getUtility(IFieldValidator, name="isValidEmail")
        self.assertEqual(v("hi@there.com"), None)
        self.assertEqual(v("one@u.washington.edu"), None)
        self.assertEqual(v(None), None)
        self.assertRaises(EmailAddressInvalid, v, "@there.com")

    def test_isCommaSeparatedEmails(self):
        v = getUtility(IFieldValidator, name="isCommaSeparatedEmails")
        self.assertEqual(v("hi@there.com,another@two.com"), None)
        self.assertEqual(v("one@u.washington.edu,  two@u.washington.edu"), None)
        self.assertEqual(v(None), None)
        self.assertNotEqual(v("abc@plone.org; xyz@plone.org"), None)


class TestCustomValidatorMessages(base.EasyFormTestCase):
    """Test friendlier validation framework"""

    def test_stringValidators(self):
        """Test string validation"""
        validators.update_validators()

        def validator(n):
            return getUtility(IFieldValidator, name=n)

        def validate(n, v):
            return validator(n) and validator(n)(v)

        self.assertRaises(ComponentLookupError, validate, "noValidator", "test")

        self.assertNotEqual(validate("isEmail", "test"), None)

        self.assertEqual(validate("isEmail", "test@test.com"), None)

        self.assertEqual(validate("isZipCode", "12345"), None)
        self.assertNotEqual(validate("isZipCode", "12345-1234"), None)


class DummyFile(object):
    def __init__(self, size=1, filename=""):
        self.size = size
        self.filename = filename

    def getSize(self):
        return self.size


classImplements(DummyFile, INamed)


class TestSizeValidator(base.EasyFormTestCase):
    def afterSetUp(self):
        self.request = self.layer["request"]
        self.validate_view = ValidateFile(self.portal, self.request)

    def test_filesize_none_validation(self):
        self.assertFalse(self.validate_view(None))

    def test_filesize_no_file_validation(self):
        self.assertFalse(self.validate_view("not a file"))

    def test_filiesize_smallsize_validation(self):
        self.assertFalse(self.validate_view(DummyFile(1024)))

    def test_filiesize_bigsize_validation(self):
        self.assertEqual(
            translate(self.validate_view(DummyFile(1000000000))),
            u"File is bigger than allowed size of 1048576 bytes!",
        )

    def test_filiesize_bigsize_custom_validation(self):
        self.assertEqual(
            translate(self.validate_view(DummyFile(1025), 1024)),
            u"File is bigger than allowed size of 1024 bytes!",
        )

    def test_forbidden_type_validation_fail(self):
        validation = self.validate_view(
            DummyFile(filename="foo.ZIP"), forbidden_types=("zip",)
        )
        self.assertEqual(translate(validation), u'File type "ZIP" is not allowed!')

    def test_forbidden_type_validation_pass(self):
        validation = self.validate_view(
            DummyFile(filename="foo.txt"), forbidden_types=("zip",)
        )
        self.assertFalse(validation)

    def test_allowed_type_validation_fail(self):
        validation = self.validate_view(
            DummyFile(filename="foo.ZIP"), allowed_types=("txt",)
        )
        self.assertEqual(translate(validation), u'File type "ZIP" is not allowed!')

    def test_allowed_type_validation_pass(self):
        validation = self.validate_view(
            DummyFile(filename="foo.txt"), allowed_types=("txt",)
        )
        self.assertFalse(validation)

    def test_allowed_type_no_ext(self):
        validation = self.validate_view(
            DummyFile(filename="foo"), allowed_types=("txt",)
        )
        self.assertEqual(translate(validation), u'File type "" is not allowed!')


@unittest.skipUnless(HAS_RECAPTCHA, "Requires plone.formwidget.recaptcha")
class TestSingleRecaptchaValidator(LoadFixtureBase):

    """Can't test captcha passes but we can test it fails"""

    schema_fixture = "recaptcha.xml"

    def afterSetUp(self):
        super(TestSingleRecaptchaValidator, self).afterSetUp()

        # Put some dummy values for recaptcha
        registry = getUtility(IRegistry)


        proxy = registry.forInterface(IReCaptchaSettings)
        proxy.public_key = u"foo"
        proxy.private_key = u"bar"

    def test_no_answer(self):
        data = {"verification": ""}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertIn("The code you entered was wrong, please enter the new one.", form)
        self.assertNotIn("Thanks for your input.", form)

    def test_wrong(self):
        data = {"verification": "123"}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertIn("The code you entered was wrong, please enter the new one.", form)
        self.assertNotIn("Thanks for your input.", form)


class TestFieldsetRecaptchaValidator(TestSingleRecaptchaValidator):
    """make sure it works inside a fieldset too"""

    schema_fixture = "fieldset_recaptcha.xml"


@unittest.skipUnless(HAS_HCAPTCHA, "Requires plone.formwidget.hcaptcha")
class TestSingleHcaptchaValidator(LoadFixtureBase):

    """Can't test captcha passes but we can test it fails
       Copy/paste test from Recaptcha, same api & add'on
       structure
    """

    schema_fixture = "hcaptcha.xml"

    def afterSetUp(self):
        super(TestSingleHcaptchaValidator, self).afterSetUp()

        # Put some dummy values for recaptcha
        registry = getUtility(IRegistry)
        proxy = registry.forInterface(IHCaptchaSettings)
        proxy.public_key = u"foo"
        proxy.private_key = u"bar"

    def test_no_answer(self):
        data = {"verification": ""}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertIn("class=\"invalid-feedback\"", form)
        self.assertNotIn("Thanks for your input.", form)

    def test_wrong(self):
        data = {"verification": "123"}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertIn("class=\"invalid-feedback\"", form)
        self.assertNotIn("Thanks for your input.", form)


class DummyUpload(FileUpload):
    def __init__(self, size, filename):
        self.file = StringIO("x" * size)
        self.file.filename = filename
        self.file.headers = []
        self.file.name = "file1"
        self.file.file = self.file
        FileUpload.__init__(self, self.file)


class TestFieldsetFileValidator(LoadFixtureBase):
    """ensure file validators works"""

    schema_fixture = "fieldset_file.xml"

    def test_wrong_type(self):
        data = {"file1": DummyUpload(20, "blah.txt")}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertNotIn("Thanks for your input.", form)
        self.assertIn('File type "TXT" is not allowed!', form)

    def test_right_type(self):
        data = {"file1": DummyUpload(20, "blah.pdf")}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertIn("Thanks for your input.", form)

    def test_too_big(self):
        data = {"file1": DummyUpload(2000, "blah.pdf")}
        request = self.LoadRequestForm(**data)
        request.method = "POST"
        form = EasyFormForm(self.ff1, request)()
        self.assertNotIn("Thanks for your input.", form)
        self.assertIn("File is bigger than allowed size of 300 bytes!", form)



# -*- coding: utf-8 -*-
#
# Integration tests for miscellaneous stuff
#

from collective.easyform.tests import base
from plone import api
from plone.namedfile.file import NamedFile
from zope.component import getMultiAdapter


class TestGetEasyFormURL(base.EasyFormTestCase):
    """test GetEasyFormURL stuff"""

    def afterSetUp(self):
        self.folder.invokeFactory("EasyForm", "ff1")
        self.form = self.folder.ff1
        self.form_url = self.form.absolute_url()

    def get_view(self, context):
        request = self.layer["request"]
        return getMultiAdapter((context, request), name="get-easyform-url")

    def check_urls_form(self, view):
        # For a view somewhere in a form
        self.assertEqual(view("fields"), self.form_url + "/fields")
        self.assertEqual(view("/fields"), self.form_url + "/fields")
        self.assertEqual(view("actions"), self.form_url + "/actions")
        self.assertEqual(view(""), self.form_url + "")

    def check_urls_non_form(self, view):
        # For a view outside a form
        self.assertEqual(view("fields"), "./fields")
        self.assertEqual(view("/fields"), "./fields")
        self.assertEqual(view("actions"), "./actions")
        self.assertEqual(view(""), "./")

    def test_get_easyform_url_form(self):
        view = self.get_view(self.form)
        self.check_urls_form(view)

    def test_get_easyform_url_fields(self):
        view = self.get_view(self.form.restrictedTraverse("fields"))
        self.check_urls_form(view)

    def test_get_easyform_url_actions(self):
        view = self.get_view(self.form.restrictedTraverse("actions"))
        self.check_urls_form(view)

    def test_get_easyform_url_folder(self):
        view = self.get_view(self.folder)
        self.check_urls_non_form(view)

    def test_get_easyform_url_portal(self):
        view = self.get_view(self.portal)
        self.check_urls_non_form(view)


class TestIsSubEasyForm(base.EasyFormTestCase):
    """test IsSubEasyForm stuff"""

    def afterSetUp(self):
        self.folder.invokeFactory("EasyForm", "ff1")
        self.form = self.folder.ff1
        self.form_url = self.form.absolute_url()

    def get_view(self, context):
        request = self.layer["request"]
        return getMultiAdapter((context, request), name="is-sub-easyform")

    def test_is_sub_easyform_form(self):
        view = self.get_view(self.form)
        self.assertFalse(view())

    def test_is_sub_easyform_fields(self):
        view = self.get_view(self.form.restrictedTraverse("fields"))
        self.assertTrue(view())

    def test_is_sub_easyform_actions(self):
        view = self.get_view(self.form.restrictedTraverse("actions"))
        self.assertTrue(view())

    def test_is_sub_easyform_folder(self):
        view = self.get_view(self.folder)
        self.assertFalse(view())

    def test_is_sub_easyform_root(self):
        view = self.get_view(self.portal)
        self.assertFalse(view())


class TestCustomTemplates(base.EasyFormTestCase):

    no_schema = u"""
  <model xmlns="http://namespaces.plone.org/supermodel/schema">
    <schema></schema>
  </model>
        """

    def test_custom_fields(self):
        default_fields = api.content.create(
            self.portal, "File", id="easyform_default_fields.xml"
        )
        default_fields.file = NamedFile(self.no_schema)
        ef = api.content.create(self.folder, "EasyForm", id="my-form")
        self.assertEqual(ef.fields_model, self.no_schema)

    def test_custom_actions(self):
        default_actions = api.content.create(
            self.portal, "File", id="easyform_default_actions.xml"
        )
        default_actions.file = NamedFile(self.no_schema)
        ef = api.content.create(self.folder, "EasyForm", id="my-form")
        self.assertEqual(ef.actions_model, self.no_schema)



# -*- coding: utf-8 -*-
from collective.easyform import easyformMessageFactory as _
from collective.easyform.browser.view import ValidateFile
from collective.easyform.interfaces import IFieldValidator
from plone import api
from Products.CMFPlone.RegistrationTool import EmailAddressInvalid
from Products.validation import validation
from Products.validation.validators.BaseValidators import baseValidators
from zope.component import provideUtility
from zope.schema import ValidationError
from z3c.form.interfaces import NOT_CHANGED
from z3c.form import validator


import six


BAD_SIGNS = frozenset(["<a ", "www.", "http:", ".com", "https:"])


def isValidEmail(value, **kwargs):
    """Check for the user email address."""
    if value is None:
        return
    reg_tool = api.portal.get_tool("portal_registration")
    if not (value and reg_tool.isValidEmail(value)):
        raise EmailAddressInvalid


def isCommaSeparatedEmails(value, **kwargs):
    """Check for one or more E-Mail Addresses separated by commas."""
    if value is None:
        # Let the system for required take care of None values
        return
    reg_tool = api.portal.get_tool("portal_registration")
    for v in value.split(","):
        if not reg_tool.isValidEmail(v.strip()):
            return _(
                u"Must be a valid list of email addresses " u"(separated by commas)."
            )


def isChecked(value, **kwargs):
    if not (
        (isinstance(value, bool) and value)
        or (isinstance(value, six.string_types) and value == "1")
    ):
        return _(u"Must be checked.")


def isUnchecked(value, **kwargs):
    if not isChecked(value):
        return _(u"Must be unchecked.")


def isNotLinkSpam(value, **kwargs):
    if not value:
        return  # No value can't be SPAM
    # validation is optional and configured on the field
    value = value.lower()
    for s in BAD_SIGNS:
        if s in value:
            return _("Links are not allowed.")


# Base validators
def update_validators():
    if validation and baseValidators:

        def method(name):
            def validate(value, **kwargs):
                if value is None:
                    # Let the system for required take care of None values
                    return
                if six.PY2 and isinstance(value, six.text_type):
                    value = value.encode("utf-8")
                res = validation(name, value, **kwargs)
                if res != 1:
                    return res

            return validate

        for validator in baseValidators:
            provideUtility(
                method(validator.name), provides=IFieldValidator, name=validator.name
            )


class FileTooBig(ValidationError):
    def __init__(self, message):
        self.message = message

    def doc(self):
        return self.message

class FileSizeValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        # The NOT_CHANGED value needs to be handled better since we may have
        # stored an invalid value in an earlier request.
        if value is NOT_CHANGED:
            value = self.widget.value
        super(FileSizeValidator, self).validate(value)
        view = ValidateFile(self.context, self.request)
        max_size = api.portal.get_registry_record("easyform.max_filesize")
        if max_size:
            result = view(value, size=max_size)
            if result:
                raise FileTooBig(result)


update_validators()



# -*- coding: utf-8 -*-
from .validators import cssClassConstraint
from .validators import isTALES
from collective.easyform import config
from collective.easyform import easyformMessageFactory as _  # NOQA
from plone.app.textfield import RichText
from plone.autoform import directives
from plone.schemaeditor.interfaces import IFieldContext
from plone.schemaeditor.interfaces import IFieldEditorExtender
from plone.schemaeditor.interfaces import ISchemaContext
from plone.schemaeditor.schema import ITextLinesField
from plone.supermodel.directives import fieldset
from plone.supermodel.model import Schema
import z3c.form
from z3c.form.interfaces import IFieldWidget
from zope import schema
from zope.component import getGlobalSiteManager
from zope.interface import Interface
from zope.interface import provider
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary

import six
import z3c.form.interfaces
import zope.interface
import zope.schema.interfaces


class WidgetVocabulary(SimpleVocabulary):
    def getTerm(self, value):
        """See zope.schema.interfaces.IBaseVocabulary"""
        if not isinstance(value, six.string_types):
            value = "{0}.{1}".format(
                value.widget_factory.__module__, value.widget_factory.__name__
            )
        return self.getTermByToken(value)


@provider(IContextSourceBinder)
def widgetsFactory(context):
    terms = []
    adapters = [
        a.factory
        for a in getGlobalSiteManager().registeredAdapters()
        if (
            a.provided == IFieldWidget
            and len(a.required) == 2
            and a.required[0].providedBy(context)
        )
    ]
    for adapter in set(adapters):
        name = u"{0}.{1}".format(adapter.__module__, adapter.__name__)
        terms.append(WidgetVocabulary.createTerm(name, str(name), adapter.__name__))
    return WidgetVocabulary(terms)


class IEasyFormFieldsEditorExtender(IFieldEditorExtender):
    pass


class IFieldExtender(Schema):
    fieldset(
        u"advanced",
        label=_("Advanced"),
        fields=["field_widget", "validators", "THidden", "depends_on", "css_class",],
    )
    directives.write_permission(field_widget=config.EDIT_ADVANCED_PERMISSION)
    field_widget = zope.schema.Choice(
        title=_(u"label_field_widget", default=u"Field Widget"),
        description=_(u"help_field_widget", default=u""),
        required=False,
        source=widgetsFactory,
    )
    directives.write_permission(validators=config.EDIT_ADVANCED_PERMISSION)
    validators = zope.schema.List(
        title=_("Validators"),
        description=_(
            u"help_userfield_validators",
            default=u"Select the validators to use on this field",
        ),
        unique=True,
        required=False,
        value_type=zope.schema.Choice(vocabulary="easyform.Validators"),
    )
    directives.write_permission(THidden=config.EDIT_TALES_PERMISSION)
    THidden = zope.schema.Bool(
        title=_(u"label_hidden", default=u"Hidden"),
        description=_(u"help_hidden", default=u"Field is hidden"),
        required=False,
        default=False,
    )
    directives.write_permission(depends_on=config.EDIT_TALES_PERMISSION)
    depends_on = zope.schema.TextLine(
        title=_(
            u'Field depends on',
        ),
        description=_(
            u'This is using the pat-depends from patternslib, all options are supported. Please read the <a href="https://patternslib.com/demos/depends" target="_blank">pat-depends documentations</a> for options.',
        ),
        default=u'',
        required=False,
    )
    directives.write_permission(css_class=config.EDIT_TALES_PERMISSION)
    css_class = zope.schema.TextLine(
        title=_(
            u'CSS Class',
        ),
        description=_(
            u'Define additional CSS class for this field here. This allowes for formating individual fields via CSS.',
        ),
        default=u'',
        required=False,
        constraint=cssClassConstraint,
    )
    fieldset(
        u"overrides",
        label=_("Overrides"),
        fields=["TDefault", "TEnabled", "TValidator", "serverSide"],
    )
    directives.write_permission(TDefault=config.EDIT_TALES_PERMISSION)
    TDefault = zope.schema.TextLine(
        title=_(u"label_tdefault_text", default=u"Default Expression"),
        description=_(
            u"help_tdefault_text",
            default=u"A TALES expression that will be evaluated when the form"
            u"is displayed to get the field default value. Leave "
            u"empty if unneeded. Your expression should evaluate as a "
            u"string. PLEASE NOTE: errors in the evaluation of this "
            u"expression will cause an error on form display.",
        ),
        default=u"",
        constraint=isTALES,
        required=False,
    )
    directives.write_permission(TEnabled=config.EDIT_TALES_PERMISSION)
    TEnabled = zope.schema.TextLine(
        title=_(u"label_tenabled_text", default=u"Enabling Expression"),
        description=_(
            u"help_tenabled_text",
            default=u"A TALES expression that will be evaluated when the form "
            u"is displayed to determine whether or not the field is "
            u"enabled. Your expression should evaluate as True if "
            u"the field should be included in the form, False if it "
            u"should be omitted. Leave this expression field empty "
            u"if unneeded: the field will be included. PLEASE NOTE: "
            u"errors in the evaluation of this expression will cause "
            u"an error on form display.",
        ),
        default=u"",
        constraint=isTALES,
        required=False,
    )
    directives.write_permission(TValidator=config.EDIT_TALES_PERMISSION)
    TValidator = zope.schema.TextLine(
        title=_(u"label_tvalidator_text", default=u"Custom Validator"),
        description=_(
            u"help_tvalidator_text",
            default=u"A TALES expression that will be evaluated when the form "
            u"is validated. Validate against 'value', which will "
            u"contain the field input. Return False if valid; if not "
            u"valid return a string error message. E.G., "
            u"\"python: test(value=='eggs', False, 'input must be "
            u'eggs\')" will require "eggs" for input. '
            u"PLEASE NOTE: errors in the evaluation of this "
            u"expression will cause an error on form display.",
        ),
        default=u"",
        constraint=isTALES,
        required=False,
    )
    directives.write_permission(serverSide=config.EDIT_TALES_PERMISSION)
    serverSide = zope.schema.Bool(
        title=_(u"label_server_side_text", default=u"Server-Side Variable"),
        description=_(
            u"description_server_side_text",
            default=u""
            u"Mark this field as a value to be injected into the "
            u"request form for use by action adapters and is not "
            u"modifiable by or exposed to the client.",
        ),
        default=False,
        required=False,
    )


class IEasyFormFieldsContext(ISchemaContext):
    """EasyForm schema view interface."""


class IEasyFormFieldContext(IFieldContext):
    """EasyForm field content marker."""


class ILabel(zope.schema.interfaces.IField):
    """Label Field."""


class IRichLabel(ILabel):
    """Rich Label Field."""

    rich_label = RichText(title=_(u"Rich Label"), default=u"", missing_value=u"")


class IEasyFormWidget(z3c.form.interfaces.IWidget):
    """General marker for easyform widgets."""


class ILabelWidget(IEasyFormWidget):
    """Label Widget."""


class IRichLabelWidget(IEasyFormWidget):
    """Rich Label Field Widget."""


class IReCaptcha(zope.schema.interfaces.ITextLine):
    """ReCaptcha Field."""


class IHCaptcha(zope.schema.interfaces.ITextLine):
    """ReCaptcha Field."""


class INorobotCaptcha(zope.schema.interfaces.ITextLine):
    """Norobot Field."""


class ILikert(zope.schema.interfaces.IField):

    questions = zope.schema.List(
        title=_(u'Possible questions'),
        description=_(u'Enter allowed choices one per line.'),
        required=zope.schema.interfaces.IChoice['vocabulary'].required,
        default=zope.schema.interfaces.IChoice['vocabulary'].default,
        value_type=zope.schema.TextLine())
    zope.interface.alsoProvides(questions, ITextLinesField)

    answers = zope.schema.List(
        title=_(u'Possible answers'),
        description=_(u'Enter allowed choices one per line.'),
        required=zope.schema.interfaces.IChoice['vocabulary'].required,
        default=zope.schema.interfaces.IChoice['vocabulary'].default,
        value_type=zope.schema.TextLine())
    zope.interface.alsoProvides(questions, ITextLinesField)

class ILikertWidget(IEasyFormWidget):
    """Likert widget."""


class IFieldValidator(Interface):
    """Base marker for field validators."""



# -*- coding: utf-8 -*-
from collections import namedtuple
from lxml import etree
from Products.PloneFormGen.content.fields import FGFieldsetEnd
from Products.PloneFormGen.content.fields import FGFieldsetStart
from Products.PloneFormGen.content.fieldsBase import BaseFormField
from Products.PloneFormGen.interfaces import IPloneFormGenFieldset

import logging
import six


logger = logging.getLogger("collective.easyform.migration")


NAMESPACES = {
    "easyform": "http://namespaces.plone.org/supermodel/easyform",
    "form": "http://namespaces.plone.org/supermodel/form",
}


def append_field(schema, type_, name, properties):
    field = etree.SubElement(schema, u"field")
    field.set(u"name", name)
    field.set(u"type", type_)
    return field


def append_label_field(schema, type_, name, properties):
    field = append_field(schema, type_, name, properties)
    append_node(field, u"required", u"False")
    return field


def append_date_field(schema, type_, name, properties):
    if properties.get("fgShowHM", False):
        return append_field(schema, u"zope.schema.Datetime", name, properties)
    else:
        return append_field(schema, u"zope.schema.Date", name, properties)


def append_fieldset(schema, type_, name, properties):
    fieldset = etree.SubElement(schema, u"fieldset")
    fieldset.set(u"name", name)
    return fieldset


def set_attribute(field, name, value):
    if u":" in name:
        ns, attr = name.split(":")
        ns = NAMESPACES.get(ns, ns)
        field.set(u"{{{}}}{}".format(ns, attr), value)
    else:
        field.set(name, value)


def append_node(field, name, value):
    if ":" in name:
        ns, name = name.split(":")
        ns = NAMESPACES.get(ns, ns)
        name = "{{{}}}{}".format(ns, name)
    node = etree.SubElement(field, name)
    if isinstance(value, (list, tuple)):
        value = u" ".join(value)
    node.text = value
    return node


def append_list_node(field, name, value):
    node = append_node(field, name, None)
    for val in value:
        el = etree.SubElement(node, "element")
        if "|" in val:
            key, val = val.split("|")
            el.set("key", key)
        el.text = val


def append_required_node(field, name, value):
    if value == u"False":
        append_node(field, name, value)


def append_maxlength_node(field, name, value):
    if value != u"0":
        append_node(field, name, value)


def append_vocab_node(field, name, value):
    if field.get("type") == "zope.schema.Set":
        node = etree.SubElement(field, "value_type")
        node.set("type", "zope.schema.Choice")
    else:
        node = field

    append_list_node(node, u"values", value)


def append_default_node(field, name, value):
    if isinstance(value, list):
        return
    if field.get("type") == "collective.easyform.fields.RichLabel":
        append_node(field, "rich_label", value)
    else:
        append_node(field, name, value)


def append_widget_node(field, name, value):
    node = append_node(field, name, None)
    type_ = field.get("type")
    if type_ == "zope.schema.Set":
        if value == "select":
            widget = "z3c.form.browser.select.CollectionSelectFieldWidget"
        else:
            widget = "z3c.form.browser.checkbox.CheckBoxFieldWidget"
    else:
        if value == "select":
            widget = "z3c.form.browser.select.ChoiceWidgetDispatcher"
        else:
            widget = "z3c.form.browser.radio.RadioFieldWidget"
    set_attribute(node, "type", widget)


def append_or_set_title(field, name, value):
    if field.tag == "fieldset":
        set_attribute(field, "label", value)
    else:
        append_node(field, name, value)


def append_or_set_description(field, name, value):
    if field.tag == 'fieldset':
        set_attribute(field, 'description', value)
    else:
        append_node(field, name, value)


def convert_tales_expressions(value):
    if value == u"here/memberEmail":
        return u"python:member and member.getProperty('email', '') or ''"
    elif value == u"here/memberFullName":
        return u"python:member and member.getProperty('fullname', '') or ''"
    elif value == u"here/memberId":
        return u"python:member and member.id or ''"
    return value


def to_text(value):
    if isinstance(value, (list, tuple)):
        return [to_text(v) for v in value]
    value = str(value)
    if six.PY2:
        value = value.decode("utf8")
    return value


Type = namedtuple("Type", ["name", "handler"])
Property = namedtuple("Property", ["name", "handler"])

TYPES_MAPPING = {
    "FormStringField": Type("zope.schema.TextLine", append_field),
    "FormPasswordField": Type("zope.schema.Password", append_field),
    "FormIntegerField": Type("zope.schema.Int", append_field),
    "FormFixedPointField": Type("zope.schema.Float", append_field),
    "FormBooleanField": Type("zope.schema.Bool", append_field),
    "FormDateField": Type("zope.schema.Date", append_date_field),
    "FormLabelField": Type("collective.easyform.fields.Label", append_label_field),
    "FormLinesField": Type("zope.schema.Text", append_field),
    "FormSelectionField": Type("zope.schema.Choice", append_field),
    "FormMultiSelectionField": Type("zope.schema.Set", append_field),
    "FormTextField": Type("zope.schema.Text", append_field),
    "FormRichTextField": Type("plone.app.textfield.RichText", append_field),
    "FormRichLabelField": Type(
        "collective.easyform.fields.RichLabel", append_label_field
    ),
    "FormFileField": Type("plone.namedfile.field.NamedBlobFile", append_field),
    "FormCaptchaField": Type("collective.easyform.fields.ReCaptcha", append_label_field),
    "FormLikertField": Type("collective.easyform.fields.Likert", append_field),
    "FieldsetStart": Type("", append_fieldset),
    "FieldsetEnd": Type("", None),
}

PROPERTIES_MAPPING = {
    "description": Property("description", append_or_set_description),
    "likertQuestions": Property("questions", append_list_node),
    "likertAnswers": Property("answers", append_list_node),
    "fgDefault": Property("default", append_default_node),
    "fgmaxlength": Property("max_length", append_maxlength_node),
    "fgsize": None,  # Not available in collective.easyform
    "fgStringValidator": Property("easyform:validators", set_attribute),
    "fgTDefault": Property("easyform:TDefault", set_attribute),
    "fgTEnabled": Property("easyform:TEnabled", set_attribute),
    "fgTValidator": Property("easyform:TValidator", set_attribute),
    "fgTVocabulary": None,  # Not available in collective.easyform
    "fgVocabulary": Property("values", append_vocab_node),
    "hidden": Property("easyform:THidden", set_attribute),
    "depends_on": Property("easyform:depends_on", set_attribute),
    "css_class": Property("easyform:css_class", set_attribute),
    "maxval": Property("max", append_node),
    "minval": Property("min", append_node),
    "placeholder": None,  # Not available in collective.easyform
    "required": Property("required", append_required_node),
    "serverSide": Property("easyform:serverSide", set_attribute),
    "title": Property("title", append_or_set_title),
    "fgFormat": Property("form:widget", append_widget_node),
}


def pfg_fields_properties(obj):
    id_ = obj.getId()
    props = {}
    props["_portal_type"] = obj.portal_type
    for field in obj.Schema().fields():
        name = field.getName()
        accessor = field.getEditAccessor(obj)
        props[name] = accessor()
    return (id_, props)


def pfg_fields(context):
    fields = []
    for obj in context.objectValues():
        if IPloneFormGenFieldset.providedBy(obj):
            # handle FieldsetFolder folderish type
            parent = obj.aq_parent
            obj_id = obj.getId()
            # append a FieldsetStart field with title and description
            start = FGFieldsetStart(obj_id).__of__(parent)
            start.setTitle(obj.Title())
            start.setDescription(obj.Description())
            fields.append(pfg_fields_properties(start))
            # add all the fieldset contained fields
            fields.extend(pfg_fields(obj))
            # finish by appending a marker FieldsetEnd field
            end = FGFieldsetEnd(obj_id).__of__(parent)
            fields.append(pfg_fields_properties(end))
            continue
        if not isinstance(obj, BaseFormField):
            continue
        fields.append(pfg_fields_properties(obj))
    return fields


def fields_model(ploneformgen):
    """Create a fields xml model from a PloneFormGen instance."""
    parser = etree.XMLParser(remove_blank_text=True)
    model = etree.fromstring(FIELDS_MODEL, parser)
    schema = model.find("{http://namespaces.plone.org/supermodel/schema}schema")
    for fieldname, properties in pfg_fields(ploneformgen):
        portal_type = properties["_portal_type"]
        if portal_type == "FieldsetEnd":
            schema = schema.getparent()
            continue

        type_ = TYPES_MAPPING.get(portal_type)
        if type_ is None:
            # ploneformgen has been renamed with _MIGRATION_ suffix
            # hence, we remove 11 chars
            pfgpath = "/".join(ploneformgen.getPhysicalPath())[:-11]
            logger.warning(
                "Ignoring field '%s' of type '%s' of form '%s'.",
                fieldname,
                portal_type,
                pfgpath,
            )
            continue

        if type_.handler is None:
            continue
        field = type_.handler(schema, type_.name, fieldname, properties)

        for name, value in properties.items():
            prop = PROPERTIES_MAPPING.get(name)
            if prop is None:
                continue

            value = to_text(value)

            # Convert TALES expressions with PFG specific methods
            if name.startswith("fgT"):
                value = convert_tales_expressions(value)

            prop.handler(field, prop.name, value)

        if portal_type == "FieldsetStart":
            schema = field

    return etree.tostring(model, pretty_print=True)


FIELDS_MODEL = b"""
<model
    xmlns="http://namespaces.plone.org/supermodel/schema"
    xmlns:easyform="http://namespaces.plone.org/supermodel/easyform"
    xmlns:i18n="http://xml.zope.org/namespaces/i18n"
    i18n:domain="collective.easyform">
  <schema></schema>
</model>
"""



# -*- coding: utf-8 -*-
from Products.CMFPlone.interfaces import INonInstallable
from zope.interface import implementer


@implementer(INonInstallable)
class HiddenProfiles(object):
    def getNonInstallableProfiles(self):
        """Hide uninstall profile from site-creation and installer."""
        return ["collective.easyform:uninstall"]

    def getNonInstallableProducts(self):
        """Hide the upgrades package from site-creation and installer.

        Our upgrades profiles are defined in the directory 'upgrades'.
        Plone sees this as a separate product.
        So instead of adding each new upgrade profile to the list of
        non installable profiles above, we can mark the upgrades product
        as non installable.
        """
        return ["collective.easyform.upgrades"]



# -*- coding: utf-8 -*-
from collective.easyform.api import get_actions
from collective.easyform.api import set_actions
from collective.easyform.tests import base
from z3c.form.interfaces import IFormLayer
from zope.interface import classImplements
from ZPublisher.BaseRequest import BaseRequest

import transaction


class FakeRequest(dict):
    def __init__(self, **kwargs):
        self.form = kwargs


class TrueOnceCalled(object):

    """A mock function that evaluates to True once it has been called."""

    def __init__(self):
        self.called = False

    def __call__(self, *args, **kw):
        self.called = True

    def __bool__(self):
        return self.called


class TestEmbedding(base.EasyFormTestCase):

    """test embedding of a PFG in another template"""

    def dummy_send(self, mfrom, mto, messageText, immediate=False):
        self.mfrom = mfrom
        self.mto = mto
        self.messageText = messageText

    def LoadRequestForm(self, **kwargs):
        form = self.app.REQUEST.form
        form.clear()
        for key in kwargs.keys():
            form[key] = kwargs[key]
        return self.app.REQUEST

    def afterSetUp(self):
        base.EasyFormTestCase.afterSetUp(self)
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")
        self.ff1.title = u"ff1"
        self.ff1.CSRFProtection = False  # no csrf protection
        actions = get_actions(self.ff1)
        actions["mailer"].recipient_email = u"mdummy@address.com"
        set_actions(self.ff1, actions)
        self.mailhost = self.folder.MailHost
        self.mailhost._send = self.dummy_send
        classImplements(BaseRequest, IFormLayer)

    def test_embedded_form_renders(self):
        view = self.ff1.restrictedTraverse("@@embedded")
        res = view()

        # form renders
        self.assertTrue("Your E-Mail Address" in res)

        # form action equals request URL
        self.assertTrue('action="{url}"'.format(url=self.ff1.absolute_url()) in res)

        # no form prefix
        # self.assertTrue('name="form.submitted"' in res)

        # we can specify a form prefix
        view.prefix = "mypfg"
        res = view()
        self.assertTrue('name="mypfg.buttons.submit"' in res)

    def test_embedded_form_validates(self):
        # fake an incomplete form submission
        self.LoadRequestForm(**{"mypfg.buttons.submit": u"Submit"})

        # render the form
        view = self.ff1.restrictedTraverse("@@embedded")
        view.prefix = "mypfg"
        res = view()

        # should stay on same page on errors, and show messages
        self.assertTrue("Required input is missing." in res)

    def test_doesnt_process_submission_of_other_form(self):
        # fake submission of a *different* form (note mismatch of form
        # submission marker with prefix)
        self.LoadRequestForm(**{"form.buttons.submit": u"Submit"})

        # let's preset a faux controller_state (as if from the other form)
        # to make sure it doesn't throw things off
        # self.app.REQUEST.set('controller_state', 'foobar')

        # render the form
        view = self.ff1.restrictedTraverse("@@embedded")
        view.prefix = "mypfg"
        res = view()

        # should be no validation errors
        self.assertFalse("Required input is missing." in res)

        # (and request should still have the 'form.submitted' key)
        # self.assertTrue('form.submitted' in self.app.REQUEST.form)

        # (and the controller state should be untouched)
        # self.assertEqual(self.app.REQUEST.get('controller_state'), 'foobar')

        # but if we remove the form prefix then it should process the form
        view.prefix = "form"
        res = view()
        self.assertTrue("Required input is missing." in res)

    def test_render_thank_you_on_success(self):
        # We need to be able to make sure the transaction commit was called
        # before the Retry exception, without actually committing our test
        # fixtures.
        real_transaction_commit = transaction.commit
        transaction.commit = TrueOnceCalled()
        # committed = TrueOnceCalled()

        self.LoadRequestForm(
            **{
                "form.widgets.topic": u"monkeys",
                "form.widgets.comments": u"I am not a walnut.",
                "form.widgets.replyto": u"foobar@example.com",
                "form.buttons.submit": u"Submit",
            }
        )
        # should raise a retry exception triggering a new publish attempt
        # with the new URL
        # XXX do a full publish for this test
        self.app.REQUEST._orig_env["PATH_TRANSLATED"] = "/plone"
        self.app.REQUEST.method = "POST"
        view = self.ff1.restrictedTraverse("@@embedded")
        # self.assertRaises(Retry, view)
        res = view()

        self.assertTrue("Thanks for your input." in res)

        # make sure the transaction was committed
        # XXX fails in python 3
        # self.assertTrue(committed)

        # make sure it can deal with VHM URLs
        self.app.REQUEST._orig_env[
            "PATH_TRANSLATED"
        ] = "/VirtualHostBase/http/nohost:80/VirtualHostRoot"
        view = self.ff1.restrictedTraverse("@@embedded")
        res = view()

        self.assertTrue("Thanks for your input." in res)
        # clean up
        transaction.commit = real_transaction_commit



# -*- coding: utf-8 -*-
from Products.CMFPlone.utils import safe_unicode

import os
import pkg_resources


this_path = os.path.dirname(__file__)


try:
    pkg_resources.get_distribution("openpyxl")
    HAS_XLSX_SUPPORT = True
except pkg_resources.DistributionNotFound:
    HAS_XLSX_SUPPORT = False


EDIT_TALES_PERMISSION = "collective.easyform.EditTALESFields"
EDIT_PYTHON_PERMISSION = "collective.easyform.EditPythonFields"
EDIT_TECHNICAL_PERMISSION = "collective.easyform.EditTechnicalFields"
EDIT_ADVANCED_PERMISSION = "collective.easyform.EditAdvancedFields"
EDIT_ADDRESSING_PERMISSION = "collective.easyform.EditMailAddresses"
USE_ENCRYPTION_PERMISSION = "collective.easyform.EditEncryptionSpecs"
DOWNLOAD_SAVED_PERMISSION = "collective.easyform.DownloadSavedInput"

with open(
    os.path.join(this_path, "default_schemata", "model_default.xml")
) as fp:  # noqa
    MODEL_DEFAULT = safe_unicode(fp.read())

with open(
    os.path.join(this_path, "default_schemata", "fields_default.xml")
) as fp:  # noqa
    FIELDS_DEFAULT = safe_unicode(fp.read())

with open(
    os.path.join(this_path, "default_schemata", "actions_default.xml")
) as fp:  # noqa
    ACTIONS_DEFAULT = safe_unicode(fp.read())

with open(
    os.path.join(this_path, "default_schemata", "mail_body_default.pt")
) as fp:  # noqa
    MAIL_BODY_DEFAULT = safe_unicode(fp.read())


FORM_ERROR_MARKER = "FORM_ERROR_MARKER"


DEFAULT_SCRIPT = u"""
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform, request
##title=
##

# Available parameters:
#  fields  = HTTP request form fields as key value pairs
#  request = The current HTTP request.
#            Access fields by request.form["myfieldname"]
#  easyform = EasyForm object
#
# Return value is not processed -- unless you
# return a dictionary with contents. That's regarded
# as an error and will stop processing of actions
# and return the user to the form. Error dictionaries
# should be of the form {'field_id':'Error message'}
# or {request.FORM_ERROR_MARKER:'Form error message'}


assert False, "Please complete your script"

"""



# -*- coding: utf-8 -*-
from Products.PageTemplates.Expressions import getEngine
from zope.tales import tales

import re
import zope.schema.interfaces


class InvalidTALESError(zope.schema.ValidationError):
    __doc__ = u"Please enter a valid TALES expression."


def isTALES(value):
    if value.strip():
        try:
            getEngine().compile(value)
        except tales.CompilerError:
            raise InvalidTALESError
    return True


class InvalidCSSClassNameError(zope.schema.ValidationError):
    __doc__ = u"Please enter valid CSS class names."


def cssClassConstraint(value):
    if not value:
        # Let the system for required take care of None values
        return True
    parts = value.strip().split(" ")
    for part in parts:
        if not re.match(r'^[A-Za-z]*[A-Za-z\-\_0-9]*[\w][\s]?$', part):
            raise InvalidCSSClassNameError
    return True




# -*- coding: utf-8 -*-
from Acquisition import aq_inner
from Acquisition import aq_parent
from collective.easyform import easyformMessageFactory as _
from collective.easyform.api import get_actions
from collective.easyform.api import get_context
from collective.easyform.api import get_schema
from collective.easyform.browser.fields import ModelEditorView
from collective.easyform.interfaces import IActionEditForm
from collective.easyform.interfaces import IActionFactory
from collective.easyform.interfaces import IEasyFormActionContext
from collective.easyform.interfaces import IEasyFormActionsContext
from collective.easyform.interfaces import IEasyFormActionsEditorExtender
from collective.easyform.interfaces import IExtraData
from collective.easyform.interfaces import INewAction
from collective.easyform.interfaces import ISaveData
from collective.easyform.interfaces import ISavedDataFormWrapper
from plone.autoform.form import AutoExtensibleForm
from plone.memoize.instance import memoize
from plone.registry.interfaces import IRegistry
from plone.schemaeditor.browser.field.traversal import FieldContext
from plone.schemaeditor.browser.schema.add_field import FieldAddForm
from plone.schemaeditor.browser.schema.listing import SchemaListing
from plone.schemaeditor.browser.schema.listing import SchemaListingPage
from plone.schemaeditor.browser.schema.traversal import SchemaContext
from plone.schemaeditor.interfaces import IFieldEditFormSchema
from plone.schemaeditor.utils import SchemaModifiedEvent
from plone.z3cform import layout
from plone.z3cform.crud import crud
from plone.z3cform.interfaces import IDeferSecurityCheck
from plone.z3cform.traversal import WrapperWidgetTraversal
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from z3c.form import field
from z3c.form import form
from zope.cachedescriptors.property import Lazy as lazy_property
from zope.component import adapter
from zope.component import getAdapters
from zope.component import getUtility
from zope.component import queryUtility
from zope.event import notify
from zope.i18nmessageid import MessageFactory
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import noLongerProvides
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema import getFieldsInOrder
from ZPublisher.BaseRequest import DefaultPublishTraverse


PMF = MessageFactory("plone")


@adapter(ISavedDataFormWrapper, IBrowserRequest)
class SavedDataTraversal(WrapperWidgetTraversal):
    def traverse(self, name, ignored):
        form = self._prepareForm()
        alsoProvides(self.request, IDeferSecurityCheck)
        form.update()
        noLongerProvides(self.request, IDeferSecurityCheck)
        for subform in form.subforms:
            if not hasattr(subform, "subforms"):
                continue
            for subsubform in subform.subforms:
                if not name.startswith(subsubform.prefix):
                    continue
                for id_ in subsubform.widgets:
                    subformname = subsubform.prefix + subsubform.widgets.prefix + id_
                    if subformname == name:
                        target = self._form_traverse(subsubform, id_)
                        target.__parent__ = aq_inner(self.context)
                        return target
        return super(SavedDataTraversal, self).traverse(name, ignored)


class SavedDataView(BrowserView):
    def items(self):
        items = [
            (name, action.__doc__)
            for name, action in getFieldsInOrder(get_actions(self.context))
            if ISaveData.providedBy(action)
        ]
        if len(items) == 1:
            url = "{0}/actions/{1}/@@data".format(
                self.context.absolute_url(),
                items[0][0],
            )
            self.request.response.redirect(url)
        return items


class DataWrapper(dict):
    def __init__(self, sid, data, parent):
        self.__sid__ = sid
        self.update(data)
        self.__parent__ = parent


class SavedDataForm(crud.CrudForm):
    template = ViewPageTemplateFile("saveddata_form.pt")
    addform_factory = crud.NullForm

    @property
    def delimiter_missing(self):
        return hasattr(self, "_delimiter_missing") and self._delimiter_missing

    @property
    def is_csv(self):
        return self.context.field.DownloadFormat == "csv"

    @property
    def csv_delimiter(self):
        if "csv_delimiter" in self.request.form:
            return self.request["csv_delimiter"]
        registry = getUtility(IRegistry)
        delimiter = registry.get("easyform.csv_delimiter").encode("utf-8")
        return delimiter

    @property
    def form_title(self):
        return aq_parent(self).Title()

    @property
    def field(self):
        return self.context.field

    @property
    def name(self):
        return self.field.__name__

    @property
    def get_schema(self):
        return get_schema(get_context(self.field))

    def description(self):
        return _(u"${items} input(s) saved", mapping={"items": self.field.itemsSaved()})

    @property
    def update_schema(self):
        fields = field.Fields(self.get_schema)
        showFields = getattr(self.field, "showFields", [])
        if showFields:
            fields = fields.select(*showFields)
        return fields

    @property
    def view_schema(self):
        ExtraData = self.field.ExtraData
        if ExtraData:
            return field.Fields(IExtraData).select(*ExtraData)


    def get_items(self):
        return [
            (key, DataWrapper(key, value, self.context))
            for key, value in self.field.getSavedFormInputItems()
        ]

    # def add(self, data):
    # storage = self.context._inputStorage

    def before_update(self, item, data):
        id_ = item.__sid__
        item.update(data)
        self.field.setDataRow(id_, item.copy())

    def remove(self, id_and_item):
        (id, item) = id_and_item
        self.field.delDataRow(id)

    @button.buttonAndHandler(PMF(u"Download"), name="download")
    def handleDownload(self, action):
        pass

    @button.buttonAndHandler(_(u"Clear all"), name="clearall")
    def handleClearAll(self, action):
        self.field.clearSavedFormInput()

    def updateActions(self):
        super(SavedDataForm, self).updateActions()
        if "download" in self.actions:
            self.actions["download"].addClass("context")

        if "clearall" in self.actions:
            self.actions["clearall"].addClass("destructive")


@implementer(ISavedDataFormWrapper)
class SavedDataFormWrapper(layout.FormWrapper):
    def __call__(self):
        if hasattr(self.request, "form.buttons.download"):
            if "csv_delimiter" in self.request.form:
                delimiter = self.request["csv_delimiter"]
                if len(delimiter) == 0:
                    self.form_instance._delimiter_missing = True
                    return super(SavedDataFormWrapper, self).__call__()
                else:
                    delimiter = delimiter[0]
                self.context.field.download(self.request.response, delimiter=delimiter)
            else:
                self.context.field.download(self.request.response)
            return u""
        if hasattr(self.context.field, 'BatchSize'):
            self.form_instance.batch_size = self.context.field.BatchSize
        return super(SavedDataFormWrapper, self).__call__()


ActionSavedDataView = layout.wrap_form(
    SavedDataForm, __wrapper_class=SavedDataFormWrapper
)


@implementer(IEasyFormActionContext)
class EasyFormActionContext(FieldContext):
    """Wrapper for published zope 3 schema fields."""

    def publishTraverse(self, request, name):
        """It's not valid to traverse to anything below a field context."""
        # hack to make inline validation work
        # (plone.app.z3cform doesn't know the form is the default view)
        if name == self.__name__:
            return ActionEditView(self, request)

        return DefaultPublishTraverse(self, request).publishTraverse(request, name)


@implementer(IEasyFormActionsContext)
class EasyFormActionsView(SchemaContext):

    schema = None

    def __init__(self, context, request):
        self.schema = get_actions(context)
        super(EasyFormActionsView, self).__init__(self.schema, request, name="actions")

    def publishTraverse(self, request, name):
        """Look up the field whose name matches the next URL path element,
        and wrap it.
        """
        try:
            return EasyFormActionContext(self.schema[name], self.request).__of__(self)
        except KeyError:
            return DefaultPublishTraverse(self, request).publishTraverse(request, name)

    def browserDefault(self, request):
        """If not traversing through the schema to a field, show the
        SchemaListingPage.
        """
        return self, ("@@listing",)


class EasyFormActionsListing(SchemaListing):
    template = ViewPageTemplateFile("actions_listing.pt")

    @memoize
    def _field_factory(self, field):
        field_identifier = u"{0}.{1}".format(field.__module__, field.__class__.__name__)
        return queryUtility(IActionFactory, name=field_identifier)

    @button.buttonAndHandler(PMF(u"Save"))
    def handleSaveDefaults(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        for fname, value in data.items():
            self.context.schema[fname].required = value
        notify(SchemaModifiedEvent(self.context))

        # update widgets to take the new defaults into account
        self.updateWidgets()
        self.request.response.redirect(aq_parent(self.context).absolute_url())

    def handleModelEdit(self, action):
        self.request.response.redirect("@@modeleditor")


class EasyFormActionsListingPage(SchemaListingPage):
    """Form wrapper so we can get a form with layout.

    We define an explicit subclass rather than using the wrap_form method
    from plone.z3cform.layout so that we can inject the schema name into
    the form label.
    """

    form = EasyFormActionsListing
    index = ViewPageTemplateFile("model_listing.pt")


class ActionAddForm(FieldAddForm):
    schema = INewAction
    fields = field.Fields()  # let fields be updated by AutoExtensibleForm.updateFields
    label = _("Add new action")


ActionAddFormPage = layout.wrap_form(ActionAddForm)


@implementer(IActionEditForm)
class ActionEditForm(AutoExtensibleForm, form.EditForm):
    def __init__(self, context, request):
        super(form.EditForm, self).__init__(context, request)
        self.field = context.field

    def getContent(self):
        return self.field

    @lazy_property
    def schema(self):
        return IFieldEditFormSchema(self.field)

    @lazy_property
    def additionalSchemata(self):
        schema_context = self.context.aq_parent
        adapters = getAdapters(
            (schema_context, self.field), IEasyFormActionsEditorExtender
        )
        return [v for k, v in adapters]

    @button.buttonAndHandler(PMF(u"Save"), name="save")
    def handleSave(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        changes = self.applyChanges(data)

        if changes:
            self.status = self.successMessage
        else:
            self.status = self.noChangesMessage

        notify(SchemaModifiedEvent(self.context.aq_parent))
        self.redirectToParent()

    @button.buttonAndHandler(PMF(u"Cancel"), name="cancel")
    def handleCancel(self, action):
        self.redirectToParent()

    def redirectToParent(self):
        parent = aq_parent(aq_inner(self.context))
        url = parent.absolute_url()
        self.request.response.redirect(url)


class ActionEditView(layout.FormWrapper):
    form = ActionEditForm

    def __init__(self, context, request):
        super(ActionEditView, self).__init__(context, request)
        self.field = context.field

    @lazy_property
    def label(self):
        return _(
            u"Edit Action '${fieldname}'", mapping={"fieldname": self.field.__name__}
        )


but = button.Button("modeleditor", title=_(u"Edit XML Actions Model"))
EasyFormActionsListing.buttons += button.Buttons(but)
handler = button.Handler(but, EasyFormActionsListing.handleModelEdit)
EasyFormActionsListing.handlers.addHandler(but, handler)


class ModelEditorView(ModelEditorView):
    """Editor view"""

    title = _(u"Edit XML Actions Model")

    def modelSource(self):
        return self.context.aq_parent.actions_model

    def save(self, source):
        self.context.aq_parent.actions_model = source



# -*- coding: utf-8 -*-
#
# Test EasyForm event-handler functionality
#

from collective.easyform.api import get_actions
from collective.easyform.api import get_schema
from collective.easyform.tests import base


class TestAdapterPaste(base.EasyFormTestCase):

    """Ensure content types can be created and edited"""

    adapterTypes = (
        "FormSaveDataAdapter",
        "FormMailerAdapter",
        "FormCustomScriptAdapter",
    )

    def afterSetUp(self):
        super(TestAdapterPaste, self).afterSetUp()
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")

    def testActiveAdaptersNotDuplicatedOnFormCopy(self):
        # self.loginAsPortalOwner()
        copy = self.folder.manage_copyObjects("ff1")
        new_id = self.folder.manage_pasteObjects(copy)[0]["new_id"]
        ff2 = getattr(self.folder, new_id)
        active_adapters = tuple(get_actions(ff2))
        self.assertEqual(active_adapters, ("mailer",))
        active_fields = tuple(get_schema(ff2))
        self.assertEqual(set(active_fields), set(["replyto", "topic", "comments"]))



# -*- coding: utf-8 -*-
from datetime import datetime, date
from plone.app.textfield import RichText, RichTextValue
from .testSaver import BaseSaveData, FakeRequest
from collective.easyform.api import get_actions, set_fields
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import IDeserializeFromJson
from zope.component import getMultiAdapter
import json
from zope.interface import Interface
from plone.supermodel import model
from plone import schema

class ISpecialConverters(model.Schema, Interface):

    datetime = schema.Datetime()
    date = schema.Date()
    set = schema.Set()
    rich = RichText()
    

class SavedDataSerializerTestCase(BaseSaveData):  

    def testDefaultFormDataSerialization(self):
        self.request = self.layer["request"]
        self.createSaver()
        saver = get_actions(self.ff1)["saver"]
        request = FakeRequest(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        saver.onSuccess(request.form, request)

        view = self.ff1.restrictedTraverse("@@actions")
        view = view.publishTraverse(view.request, "saver")
        view = view.publishTraverse(view.request, "data")
        view.update()
        form = view.form_instance
        message = form.description()
        self.assertEqual(message.mapping, {"items": 1})
        item = form.get_items()[0]
        self.assertEqual(item[1]["id"], item[0])
        self.assertEqual(item[1]["topic"], "test subject")
        self.assertEqual(item[1]["replyto"], "test@test.org")
        self.assertEqual(item[1]["comments"], "test comments")
        obj = self.serialize(self.ff1)
        self.assertIn("savedDataStorage", obj)
        self.assertIn("test subject", json.dumps(obj))
        self.assertIn("test@test.org", json.dumps(obj))
        self.assertIn("test comments", json.dumps(obj))

    def testSpecialConverterFormDataSerialization(self):
        self.request = self.layer["request"]
        set_fields(self.ff1, ISpecialConverters)
        self.createSaver()
        saver = get_actions(self.ff1)["saver"]
        request = FakeRequest(
            datetime=datetime(1999, 12, 25, 19, 0, 0),
            date=date(2004, 11, 30),
            set=set(["please", "kill", "me"]),
            rich=RichTextValue(raw=u"<div><b>testing is fùn</b> says Michaël</div>"),
        )
        saver.onSuccess(request.form, request)

        view = self.ff1.restrictedTraverse("@@actions")
        view = view.publishTraverse(view.request, "saver")
        obj = self.serialize(self.ff1)
        self.assertIn("savedDataStorage", obj)
        self.assertIn("1999-12-25T19:00:00", json.dumps(obj))
        self.assertIn("2004-11-30", json.dumps(obj))
        self.assertIn("please", json.dumps(obj))
        self.assertIn("kill", json.dumps(obj))
        self.assertIn("me", json.dumps(obj))
        self.assertIn(u"<div><b>testing is f\\u00f9n</b> says Micha\\u00ebl</div>", json.dumps(obj))

    def serialize(self, obj=None):
        if obj is None:
            obj = self.portal.doc1
        serializer = getMultiAdapter((obj, self.request), ISerializeToJson)
        return serializer()


class SavedDataDeserializerTestCase(BaseSaveData):

    def testDefaultFormDataDeserialization(self):
        BODY = ( '{"actions_model": "<model '
                'xmlns:i18n=\\"http://xml.zope.org/namespaces/i18n\\" '
                'xmlns:marshal=\\"http://namespaces.plone.org/supermodel/marshal\\" '
                'xmlns:form=\\"http://namespaces.plone.org/supermodel/form\\" '
                'xmlns:security=\\"http://namespaces.plone.org/supermodel/security\\" '
                'xmlns:users=\\"http://namespaces.plone.org/supermodel/users\\" '
                'xmlns:lingua=\\"http://namespaces.plone.org/supermodel/lingua\\" '
                'xmlns:easyform=\\"http://namespaces.plone.org/supermodel/easyform\\" '
                'xmlns=\\"http://namespaces.plone.org/supermodel/schema\\" '
                'i18n:domain=\\"collective.easyform\\">\\n  <schema>\\n   <field '
                'name=\\"saver\\" type=\\"collective.easyform.actions.SaveData\\">\\n      '
                '<title>Saver</title>\\n    </field>\\n  </schema>\\n</model>",'
                '"savedDataStorage": '
                '{"saver": {"1658237759974": {"topic": "test subject", "replyto": '
                '"test@test.org", "comments": "test comments", "id": 1658237759974}}}}')
        self.deserialize(body=BODY, context=self.ff1)
        saver = get_actions(self.ff1)["saver"]
        data = saver.getSavedFormInput()[0]
        self.assertIn("topic", data)
        self.assertEqual("test subject", data['topic'])
        self.assertIn("replyto", data)
        self.assertEqual("test@test.org", data['replyto'])
        self.assertIn("comments", data)
        self.assertEqual("test comments", data['comments'])

    def testSpecialConvertersFormDataDeserialization(self):
        BODY = ( '{"fields_model": "<model xmlns:i18n=\\"http://xml.zope.org/namespaces/i18n\\" '
                'xmlns:marshal=\\"http://namespaces.plone.org/supermodel/marshal\\" '
                'xmlns:form=\\"http://namespaces.plone.org/supermodel/form\\" '
                'xmlns:security=\\"http://namespaces.plone.org/supermodel/security\\" '
                'xmlns:users=\\"http://namespaces.plone.org/supermodel/users\\" '
                'xmlns:lingua=\\"http://namespaces.plone.org/supermodel/lingua\\" '
                'xmlns:easyform=\\"http://namespaces.plone.org/supermodel/easyform\\" '
                'xmlns=\\"http://namespaces.plone.org/supermodel/schema\\">\\n  <schema '
                'based-on=\\"zope.interface.Interface\\">\\n    <field name=\\"datetime\\" '
                'type=\\"zope.schema.Datetime\\"/>\\n    <field name=\\"date\\" '
                'type=\\"zope.schema.Date\\"/>\\n    <field name=\\"set\\" '
                'type=\\"zope.schema.Set\\"/>\\n    <field name=\\"rich\\" '
                'type=\\"plone.app.textfield.RichText\\"/>\\n  </schema>\\n</model>", '
                '"actions_model": "<model '
                'xmlns:i18n=\\"http://xml.zope.org/namespaces/i18n\\" '
                'xmlns:marshal=\\"http://namespaces.plone.org/supermodel/marshal\\" '
                'xmlns:form=\\"http://namespaces.plone.org/supermodel/form\\" '
                'xmlns:security=\\"http://namespaces.plone.org/supermodel/security\\" '
                'xmlns:users=\\"http://namespaces.plone.org/supermodel/users\\" '
                'xmlns:lingua=\\"http://namespaces.plone.org/supermodel/lingua\\" '
                'xmlns:easyform=\\"http://namespaces.plone.org/supermodel/easyform\\" '
                'xmlns=\\"http://namespaces.plone.org/supermodel/schema\\" '
                'i18n:domain=\\"collective.easyform\\">\\n  <schema>\\n   <field '
                'name=\\"saver\\" type=\\"collective.easyform.actions.SaveData\\">\\n      '
                '<title>Saver</title>\\n    </field>\\n  </schema>\\n</model>",'
                '"savedDataStorage": '
                '{"saver": {"1658240583228": {"datetime": "1999-12-25T19:00:00", "date": '
                '"2004-11-30", "set": ["please", "me", "kill"], "rich": "<div><b>testing is '
                'f\\u00f9n</b> says Micha\\u00ebl</div>", "id": 1658240583228}}}}')

        self.deserialize(body=BODY, context=self.ff1)
        saver = get_actions(self.ff1)["saver"]
        data = saver.getSavedFormInput()[0]
        self.assertIn("datetime", data)
        self.assertEqual(datetime(1999, 12, 25, 19, 0, 0), data['datetime'])
        self.assertIn("date", data)
        self.assertEqual(datetime(2004, 11, 30), data['date'])
        self.assertIn("set", data)
        setdata = list(data['set'])
        setdata.sort()
        self.assertEqual(["kill", "me", "please"], setdata)
        self.assertIn("rich", data)
        self.assertEqual(u"<div><b>testing is fùn</b> says Michaël</div>", data['rich'].raw)

    
    def deserialize(self, body="{}", validate_all=False, context=None, create=False):
        self.request = self.layer["request"]
        self.request["BODY"] = body
        deserializer = getMultiAdapter((context, self.request), IDeserializeFromJson)
        return deserializer(validate_all=validate_all, create=create)


# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


version = "4.3.0"

setup(
    name="collective.easyform",
    version=version,
    description="Forms for Plone",
    long_description=open("README.rst").read() + "\n" + open("CHANGES.rst").read(),
    # Get more strings from https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: Addon",
        "Framework :: Plone :: 6.0",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    ],
    keywords="form plone ttw builder email",
    author="Roman Kozlovskyi",
    author_email="krzroman@gmail.com",
    maintainer="Plone Release Team",
    maintainer_email="releaseteam@plone.org",
    url="https://github.com/collective/collective.easyform",
    license="GPLv2",
    packages=find_packages("src", exclude=["ez_setup"]),
    namespace_packages=["collective"],
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "plone.api",
        "plone.app.dexterity",
        "plone.app.textfield >= 1.2.8",
        "plone.autoform",
        "plone.dexterity",
        "plone.namedfile",
        "plone.schema",
        "plone.schemaeditor>=4.0.0b1",
        "plone.supermodel",
        "plone.restapi",
        "Products.CMFPlone>=6.0.0b1",
        "Products.validation>=3.0.0",
        "setuptools",
        "six",
        # -*- Extra requirements: -*-
    ],
    extras_require={
        "downloadxlsx": ["openpyxl"],
        "recaptcha": ["plone.formwidget.recaptcha"],
        "hcaptcha": ["plone.formwidget.hcaptcha"],
        "norobots": ["collective.z3cform.norobots"],
        "ploneformgen": [
            "Products.PloneFormGen",
        ],
        "test": [
            "beautifulsoup4",
            "plone.app.testing[robot]",
            "plone.app.robotframework",
            "plone.app.contenttypes",
            "plone.formwidget.recaptcha",
            "plone.formwidget.hcaptcha",
            "robotframework-selenium2library",
            "robotframework-selenium2screenshots",
        ],
    },
    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



# -*- coding: utf-8 -*-
from collective.easyform.tests.base import FUNCTIONAL_TESTING
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.testing import layered
from plone.testing.zope import Browser

import doctest
import os
import re
import six
import unittest


LINESEP = b"\r\n"

optionflags = (
    doctest.REPORT_ONLY_FIRST_FAILURE
    | doctest.NORMALIZE_WHITESPACE
    | doctest.ELLIPSIS
    | doctest.REPORTING_FLAGS
)

testfiles = (
    "../api.py",
    "attachment.rst",
    "basic.rst",
    "browser.rst",
    "serverside_field.rst",
    "ssl.rst",
    "action_errors.rst",
)


class Py23DocChecker(doctest.OutputChecker):
    def check_output(self, want, got, optionflags):
        if six.PY2:
            got = re.sub("zExceptions.NotFound", "NotFound", got)
            got = re.sub("u'(.*?)'", "'\\1'", want)
        return doctest.OutputChecker.check_output(self, want, got, optionflags)


def get_browser(layer, auth=True):
    browser = Browser(layer["app"])
    browser.handleErrors = False
    if auth:
        browser.open('http://nohost/plone/login_form')
        browser.getControl('Login Name').value = SITE_OWNER_NAME
        browser.getControl('Password').value = SITE_OWNER_PASSWORD
        browser.getControl('Log in').click()
    return browser


def get_image_path():
    dir_name = os.path.dirname(os.path.realpath(__file__))
    return "{0}/PloneLogo.png".format(dir_name)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTests(
        [
            layered(
                doctest.DocFileSuite(
                    f,
                    optionflags=optionflags,
                    globs={
                        "LINESEP": LINESEP,
                        "get_browser": get_browser,
                        "get_image_path": get_image_path,
                    },
                    checker=Py23DocChecker(),
                ),
                layer=FUNCTIONAL_TESTING,
            )
            for f in testfiles
        ]
    )
    return suite
